"""Testing fmu-ensemble, how it works on ensembles with realizations where Eclipse
has stopped preamturely

In the test data directory, there is a file 2_R001_REEK-1.UNSMRY-failed2000
which is the result of the same DATA file but with an END keyword injected
in the Schedule section, to simulate a premature failure. FMU-ensemble does
not try to distuingish between early exits from failure or deliberate (like this).
"""

import datetime
import logging
import os
import shutil

import numpy as np
import pandas as pd
from fmu.ensemble import ScratchEnsemble

logger = logging.getLogger(__name__)


def test_ens_premature_ecl(tmpdir):
    """Check an ensemble where Eclipse has failed early in realization 1"""
    if "__file__" in globals():
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    origensemble = ScratchEnsemble(
        "origreek", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    raw_orig_smry = origensemble.load_smry()
    # Copy the ensemble to /tmp so we can modify the UNSMRY file in real 2:
    tmpdir.chdir()

    shutil.copytree(testdir + "/data/testensemble-reek001", "ens_fail_real_reek001")
    unsmry_filename = (
        "ens_fail_real_reek001/realization-1/"
        + "iter-0/eclipse/model/2_R001_REEK-1.UNSMRY"
    )
    shutil.copy(unsmry_filename + "-failed2000", unsmry_filename)

    failensemble = ScratchEnsemble(
        "failedreek", "ens_fail_real_reek001/realization-*/iter-0"
    )
    raw_fail_smry = failensemble.load_smry()

    # This is usually superfluous when raw datetimes are obtained.
    raw_orig_smry["DATE"] = pd.to_datetime(raw_orig_smry["DATE"])
    raw_fail_smry["DATE"] = pd.to_datetime(raw_fail_smry["DATE"])

    # Homogeneous max-date in orig smry:
    assert len(raw_orig_smry.groupby("REAL").max()["DATE"].unique()) == 1
    # Different values for raw_fail:
    assert len(raw_fail_smry.groupby("REAL").max()["DATE"].unique()) == 2
    # END statement in schedule file on 2000-08-01 yields this:
    assert (
        str(raw_fail_smry.groupby("REAL").max()["DATE"].loc[1]) == "2000-08-01 00:00:00"
    )

    # Filter away all those that did not make it to the end. In normal scenarios,
    # this would be accomplished by .filter('OK'), but not in this test context.
    max_date = str(failensemble.get_smry()["DATE"].max())
    filtered_fail_ensemble = failensemble.filter(
        "unsmry--raw", column="DATE", columncontains=max_date, inplace=False
    )
    assert len(filtered_fail_ensemble) == 4
    assert (
        len(filtered_fail_ensemble.get_smry().groupby("REAL").max()["DATE"].unique())
        == 1
    )
    # Check also get_smry():
    assert len(failensemble.get_smry().groupby("REAL").max()["DATE"].unique()) == 2

    # With time_index set to something, then all realization will get
    # interpolated onto the same date range
    assert (
        len(
            failensemble.get_smry(time_index="monthly")
            .groupby("REAL")
            .max()["DATE"]
            .unique()
        )
        == 1
    )
    # This is in fact *different* from what you would get from load_smry (issue #97)
    assert (
        len(
            failensemble.load_smry(time_index="monthly")
            .groupby("REAL")
            .max()["DATE"]
            .unique()
        )
        == 2
    )
    # (this behaviour might change, get_smry() is allowed in
    # the future to mimic load_smry())

    # Check that FOPT is very much lower in real 1 in failed ensemble:
    assert (
        failensemble.get_smry(column_keys="FOPT", time_index="monthly")
        .groupby("REAL")
        .max()["FOPT"]
        .loc[1]
        < 1500000
    )
    assert (
        origensemble.get_smry(column_keys="FOPT", time_index="monthly")
        .groupby("REAL")
        .max()["FOPT"]
        .loc[1]
        > 6000000
    )

    # Also for yearly
    assert (
        failensemble.get_smry(column_keys="FOPT", time_index="yearly")
        .groupby("REAL")
        .max()["FOPT"]
        .loc[1]
        < 1500000
    )
    assert (
        origensemble.get_smry(column_keys="FOPT", time_index="yearly")
        .groupby("REAL")
        .max()["FOPT"]
        .loc[1]
        > 6000000
    )

    fail_foprs = failensemble.get_smry(column_keys="FOPR", time_index="monthly")

    # The FOPR rate vector should be all zero after the stop
    assert (
        fail_foprs[
            (fail_foprs["REAL"] == 1) & (fail_foprs["DATE"] > datetime.date(2000, 8, 1))
        ]["FOPR"]
        .abs()
        .sum()
        == 0
    )
    assert (
        fail_foprs[
            (fail_foprs["REAL"] == 0) & (fail_foprs["DATE"] > datetime.date(2000, 8, 1))
        ]["FOPR"]
        .abs()
        .sum()
        > 0
    )

    # This frame treats the "failed" realization as correct,
    # and it will affect the stats:
    fail_stats = failensemble.get_smry_stats(time_index="monthly")
    # Here, real 1 is removed
    filtered_stats = filtered_fail_ensemble.get_smry_stats(time_index="monthly")
    # Original stats
    orig_stats = origensemble.get_smry_stats(time_index="monthly")

    # The 30 last rows are the rows from 2000-09-01 to 2003-02-01:
    assert fail_stats.loc["minimum"]["FOPR"].iloc[-30:].abs().sum() == 0
    assert fail_stats.loc["minimum"]["FOPT"].iloc[-30:].unique()[0] == 1431247.125
    # Oh no, in filtered stats, the last date 2003-02-01 is
    # not included, probably a minor bug!
    # But that means that the indexing of the last 30 is a little bit rogue.
    # (this test should work even that bug is fixed)
    assert filtered_stats.loc["minimum"]["FOPR"].iloc[-29:].abs().sum() > 0
    assert len(filtered_stats.loc["minimum"]["FOPT"].iloc[-29:].unique()) == 29

    # Mean FOPR and FOPT should be affected by the zero-padded rates:
    assert (
        fail_stats.loc["mean"].iloc[-10]["FOPR"]
        < filtered_stats.loc["mean"].iloc[-10]["FOPR"]
    )
    assert (
        fail_stats.loc["mean"].iloc[-10]["FOPR"]
        < orig_stats.loc["mean"].iloc[-10]["FOPR"]
    )
    assert (
        fail_stats.loc["mean"].iloc[-10]["FOPT"]
        < filtered_stats.loc["mean"].iloc[-10]["FOPT"]
    )
    assert (
        fail_stats.loc["mean"].iloc[-10]["FOPT"]
        < orig_stats.loc["mean"].iloc[-10]["FOPT"]
    )

    # Delta profiles:
    delta_fail = origensemble - failensemble
    # Delta profiles are given for all realizations
    delta_fail_smry = delta_fail.get_smry()
    assert len(delta_fail_smry["REAL"].unique()) == 5
    # and they all end at the same ultimate date:
    assert len(delta_fail_smry.groupby("REAL").max()["DATE"].unique()) == 1
    # BUT, there is only NaNs for values after 2000-08-01:
    assert np.isnan(
        delta_fail_smry[
            (delta_fail_smry["REAL"] == 1) & (delta_fail_smry["DATE"] > "2000-08-01")
        ]["FOPT"].unique()[0]
    )

    # Delta profiles after filtering:
    delta_filtered = origensemble - filtered_fail_ensemble
    assert len(origensemble) == 5
    assert len(filtered_fail_ensemble) == 4
    # assert len(delta_filtered) == 4  # Only four realizations (requires #83 resolved)
    # to_virtual() and time_index can be removed when #83 is finished.
    delta_filtered_smry = delta_filtered.to_virtual().get_smry(time_index="monthly")
    # Should contain only four realizations, as one has been filtered away
    assert len(delta_filtered_smry["REAL"].unique()) == 4
    # Ultimate date is the same in all four:
    assert len(delta_filtered_smry.groupby("REAL").max()["DATE"].unique()) == 1



from datetime import datetime as dt

import pytest
from fmu.ensemble.util.dates import _fallback_date_roll, date_range

# These tests are duplicated from https://github.com/equinor/res2df/blob/master/tests/test_summary.py


@pytest.mark.parametrize(
    "rollme, direction, freq, expected",
    [
        (
            dt(3000, 1, 1),
            "forward",
            "yearly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 1, 1),
            "forward",
            "monthly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 1, 2),
            "forward",
            "yearly",
            dt(3001, 1, 1),
        ),
        (
            dt(3000, 1, 2),
            "forward",
            "monthly",
            dt(3000, 2, 1),
        ),
        (
            dt(3000, 1, 1),
            "back",
            "yearly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 1, 1),
            "back",
            "monthly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 12, 31),
            "back",
            "yearly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 2, 2),
            "back",
            "monthly",
            dt(3000, 2, 1),
        ),
        pytest.param(
            dt(3000, 2, 2),
            "forward",
            "daily",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            dt(3000, 2, 2),
            "upwards",
            "yearly",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_fallback_date_roll(rollme, direction, freq, expected):
    """The pandas date rolling does not always work for years beyound 2262. The
    code should fallback automatically to hide that Pandas limitation"""
    assert _fallback_date_roll(rollme, direction, freq) == expected


@pytest.mark.parametrize(
    "start, end, freq, expected",
    [
        (
            dt(3000, 1, 1),
            dt(3002, 1, 1),
            "yearly",
            [
                dt(3000, 1, 1),
                dt(3001, 1, 1),
                dt(3002, 1, 1),
            ],
        ),
        (
            dt(2999, 11, 1),
            dt(3000, 2, 1),
            "monthly",
            [
                dt(2999, 11, 1),
                dt(2999, 12, 1),
                dt(3000, 1, 1),
                dt(3000, 2, 1),
            ],
        ),
        pytest.param(
            dt(3000, 1, 1),
            dt(3000, 2, 1),
            "weekly",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        (
            # Crossing the problematic time boundary:
            dt(2260, 1, 1),
            dt(2263, 1, 1),
            "yearly",
            [
                dt(2260, 1, 1),
                dt(2261, 1, 1),
                dt(2262, 1, 1),
                dt(2263, 1, 1),
            ],
        ),
        (
            dt(3000, 1, 1),
            dt(3000, 1, 1),
            "yearly",
            [
                dt(3000, 1, 1),
            ],
        ),
        (
            dt(2000, 1, 1),
            dt(2000, 1, 1),
            "yearly",
            [
                dt(2000, 1, 1),
            ],
        ),
        (
            dt(2000, 1, 1),
            dt(1000, 1, 1),
            "yearly",
            [],
        ),
        (
            dt(3000, 1, 1),
            dt(2000, 1, 1),
            "yearly",
            [],
        ),
        (
            dt(2300, 5, 6),
            dt(2302, 3, 1),
            "yearly",
            [
                dt(2300, 5, 6),
                dt(2301, 1, 1),
                dt(2302, 1, 1),
                dt(2302, 3, 1),
            ],
        ),
        (
            dt(2304, 5, 6),
            dt(2302, 3, 1),
            "yearly",
            [],
        ),
        (
            dt(2302, 3, 1),
            dt(2302, 3, 1),
            "yearly",
            [dt(2302, 3, 1)],
        ),
    ],
)
def test_date_range(start, end, freq, expected):
    """When dates are beyond year 2262,
    the function _fallback_date_range() is triggered."""
    assert date_range(start, end, freq) == expected



"""Testing observations in fmu-ensemble."""

import datetime
import glob
import logging
import os

import dateutil
import numpy as np
import pandas as pd
import pytest
import yaml
from fmu.ensemble import EnsembleSet, Observations, ScratchEnsemble, ScratchRealization

logger = logging.getLogger(__name__)


def test_observation_import(tmpdir):
    """Test import of observations from yaml"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    obs = Observations(
        testdir
        + "/data/testensemble-reek001/"
        + "/share/observations/"
        + "observations.yml"
    )
    assert len(obs.keys()) == 2  # adjust this..
    assert len(obs["smry"]) == 7
    assert len(obs["rft"]) == 2

    assert isinstance(obs["smry"], list)
    assert isinstance(obs["rft"], list)

    # Dump back to disk
    tmpdir.chdir()
    exportedfile = "share/observations/observations_copy.yml"
    obs.to_disk(exportedfile)
    assert os.path.exists(exportedfile)


def test_real_mismatch():
    """Test calculation of mismatch from the observation set to a
    realization"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    real = ScratchRealization(
        testdir + "/data/testensemble-reek001/" + "realization-0/iter-0/"
    )

    real.load_smry()
    real.load_txt("outputs.txt")
    real.load_scalar("npv.txt")

    obs = Observations(
        {"txt": [{"localpath": "parameters.txt", "key": "FWL", "value": 1702}]}
    )
    realmis = obs.mismatch(real)

    # Check layout of returned data
    assert isinstance(realmis, pd.DataFrame)
    assert len(realmis) == 1
    assert "REAL" not in realmis.columns  # should only be there for ensembles.
    assert "OBSTYPE" in realmis.columns
    assert "OBSKEY" in realmis.columns
    assert "DATE" not in realmis.columns  # date is not relevant
    assert "LABEL" in realmis.columns
    assert "MISMATCH" in realmis.columns
    assert "L1" in realmis.columns
    assert "L2" in realmis.columns

    # Check actually computed values, there should only be one row with data:
    assert realmis.loc[0, "OBSTYPE"] == "txt"
    assert realmis.loc[0, "OBSKEY"] == "parameters.txt/FWL"
    assert set(realmis["LABEL"]) == {""}
    assert realmis.loc[0, "MISMATCH"] == -2
    assert realmis.loc[0, "SIGN"] == -1
    assert realmis.loc[0, "L1"] == 2
    assert realmis.loc[0, "L2"] == 4

    # Another observation set:
    obs2 = Observations(
        {
            "txt": [
                {"localpath": "parameters.txt", "key": "RMS_SEED", "value": 600000000},
                {"localpath": "outputs.txt", "key": "top_structure", "value": 3200},
            ],
            "scalar": [{"key": "npv.txt", "value": 3400}],
        }
    )
    realmis2 = obs2.mismatch(real)
    assert len(realmis2) == 3
    assert "parameters.txt/RMS_SEED" in realmis2["OBSKEY"].values
    assert "outputs.txt/top_structure" in realmis2["OBSKEY"].values
    assert "npv.txt" in realmis2["OBSKEY"].values

    # assert much more!

    # Test that we can write the observations to yaml
    # and verify that the exported yaml can be reimported
    # and yield the same result
    obs2r = Observations(yaml.full_load(obs2.to_yaml()))
    realmis2r = obs2r.mismatch(real)
    assert np.all(
        realmis2["MISMATCH"].values.sort() == realmis2r["MISMATCH"].values.sort()
    )

    # Test use of allocated values:
    obs3 = Observations({"smryh": [{"key": "FOPT", "histvec": "FOPTH"}]})
    fopt_mis = obs3.mismatch(real)
    assert fopt_mis.loc[0, "OBSTYPE"] == "smryh"
    assert fopt_mis.loc[0, "OBSKEY"] == "FOPT"
    assert fopt_mis.loc[0, "L1"] > 0
    assert fopt_mis.loc[0, "L1"] != fopt_mis.loc[0, "L2"]

    # Test mismatch where some data is missing:
    obs4 = Observations({"smryh": [{"key": "FOOBAR", "histvec": "FOOBARH"}]})
    mis_mis = obs4.mismatch(real)
    assert mis_mis.empty

    # This test fails, the consistency check is not implemented.
    # obs_bogus = Observations({'smryh': [{'keddy': 'FOOBAR',
    #                               'histdddvec': 'FOOBARH'}]})
    # mis_mis = obs_bogus.mismatch(real)
    # assert mis_mis.empty

    obs_bogus_scalar = Observations(
        {"scalar": [{"key": "nonexistingnpv.txt", "value": 3400}]}
    )
    # (a warning should be logged)
    assert obs_bogus_scalar.mismatch(real).empty

    obs_bogus_param = Observations(
        {
            "txt": [
                {
                    "localpath": "bogusparameters.txt",
                    "key": "RMS_SEED",
                    "value": 600000000,
                }
            ]
        }
    )
    # (a warning should be logged)
    assert obs_bogus_param.mismatch(real).empty

    obs_bogus_param = Observations(
        {
            "txt": [
                {
                    "localpath": "parameters.txt",
                    "key": "RMS_SEEEEEEED",
                    "value": 600000000,
                }
            ]
        }
    )
    # (a warning should be logged)
    assert obs_bogus_param.mismatch(real).empty

    # Non-existing summary key:
    obs_bogus_smry = Observations(
        {
            "smry": [
                {
                    "key": "WBP4:OP_XXXXX",
                    "observations": [
                        {"date": datetime.date(2001, 1, 1), "error": 4, "value": 251}
                    ],
                }
            ]
        }
    )
    assert obs_bogus_smry.mismatch(real).empty

    # Test dumping to yaml:
    # Not implemented.
    # yamlobsstr = obs2.to_yaml()
    # assert isinstance(yamlobsstr, str)
    # * Write yamlobsstr to tmp file
    # * Reload observation object from that file
    # * Check that the observation objects are the same


def test_smry():
    """Test the support for smry observations, these are
    observations relating to summary data, but where
    the observed values are specified in yaml, not through
    *H summary variables"""

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    obs = Observations(
        testdir
        + "/data/testensemble-reek001/"
        + "/share/observations/"
        + "observations.yml"
    )
    real = ScratchRealization(
        testdir + "/data/testensemble-reek001/" + "realization-0/iter-0/"
    )

    # Compute the mismatch from this particular observation set to the
    # loaded realization.
    mismatch = obs.mismatch(real)

    assert len(mismatch) == 21  # later: implement counting in the obs object
    assert mismatch.L1.sum() > 0
    assert mismatch.L2.sum() > 0

    # This should work, but either the observation object
    # must do the smry interpolation in dataframes, or
    # the virtual realization should implement get_smry()
    # vreal = real.to_virtual()
    # vmismatch = obs.mismatch(vreal)
    # print(vmismatch)


def test_smry_labels():
    testdir = os.path.dirname(os.path.abspath(__file__))
    real = ScratchRealization(
        testdir + "/data/testensemble-reek001/" + "realization-0/iter-0/"
    )
    real.load_smry()
    obs_pr = Observations(
        {
            "smry": [
                {
                    "key": "WBP4:OP_1",
                    "comment": "Pressure observations well OP_1",
                    "observations": [
                        {
                            "value": 250,
                            "error": 1,
                            "date": datetime.date(2001, 1, 1),
                            "label": "WBP4_OP_1_01",
                        }
                    ],
                }
            ]
        }
    )
    mismatch = obs_pr.mismatch(real)
    assert mismatch["LABEL"].values == ["WBP4_OP_1_01"]


def test_errormessages():
    """Test that we give ~sensible error messages when the
    observation input is unparseable"""

    # Emtpy
    with pytest.raises(TypeError):
        # pylint: disable=E1120
        Observations()

    # Non-existing filename:
    with pytest.raises(IOError):
        Observations("foobar")

    # Integer input does not make sense..
    with pytest.raises(ValueError):
        Observations(3)

    # Unsupported observation category, this foobar will be wiped
    emptyobs = Observations(dict(foobar="foo"))
    assert emptyobs.empty
    # (there will be logged a warning)

    # Empty observation set should be ok, but it must be a dict
    empty2 = Observations(dict())
    assert empty2.empty
    with pytest.raises(ValueError):
        Observations([])

    # Check that the dict is a dict of lists:
    assert Observations(dict(smry="not_a_list")).empty
    # (warning will be printed)

    # This should give a warning because 'observation' is missing
    wrongobs = Observations(
        {"smry": [{"key": "WBP4:OP_1", "comment": "Pressure observations well OP_1"}]}
    )
    assert wrongobs.empty


def test_smryh():
    """Test that smryh mismatch calculation will respect time index"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    ens = ScratchEnsemble(
        "test", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/"
    )

    obs_yearly = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "yearly"}]}
    )
    obs_raw = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "raw"}]}
    )
    obs_monthly = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "monthly"}]}
    )
    obs_daily = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "daily"}]}
    )
    obs_last = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "last"}]}
    )
    obs_isodatestr = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "2003-02-01"}]}
    )
    obs_future = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "3003-02-01"}]}
    )
    obs_past = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "1003-02-01"}]}
    )

    assert obs_isodatestr
    obs_isodate = Observations(
        {
            "smryh": [
                {
                    "key": "FOPT",
                    "histvec": "FOPTH",
                    "time_index": dateutil.parser.isoparse("2003-02-01"),
                }
            ]
        }
    )
    assert obs_isodate

    obs_error = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "ølasjkdf"}]}
    )
    assert not obs_error
    obs_error2 = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": 4.43}]}
    )
    assert not obs_error2

    mismatchyearly = obs_yearly.mismatch(ens)
    mismatchmonthly = obs_monthly.mismatch(ens)
    mismatchdaily = obs_daily.mismatch(ens)
    mismatchlast = obs_last.mismatch(ens)
    mismatchraw = obs_raw.mismatch(ens)
    assert mismatchraw["TIME_INDEX"].unique() == ["raw"]

    mismatchdate = obs_isodate.mismatch(ens)
    assert "2003-02-01" in mismatchdate["TIME_INDEX"].unique()[0]

    mismatchdatestr = obs_isodatestr.mismatch(ens)
    # There might be a clock time included
    assert "2003-02-01" in mismatchdatestr["TIME_INDEX"].unique()[0]
    assert all(mismatchdate["L1"] == mismatchdatestr["L1"])

    mismatchfuture = obs_future.mismatch(ens)
    assert all(mismatchfuture["L1"] == mismatchlast["L1"])

    mismatchpast = obs_past.mismatch(ens)
    assert np.isclose(sum(mismatchpast["L2"]), 0.0)

    # When only one datapoint is included, these should be identical:
    assert (mismatchlast["L1"] == mismatchlast["L2"]).all()
    assert (mismatchlast["L1"] == mismatchlast["MISMATCH"].abs()).all()

    # Check that we have indeed calculated things differently between the time indices:
    assert mismatchyearly["L2"].sum != mismatchmonthly["L2"].sum()
    assert mismatchdaily["L2"].sum != mismatchraw["L2"].sum()


def test_ens_mismatch():
    """Test calculation of mismatch to ensemble data"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")
    ens = ScratchEnsemble(
        "test", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/"
    )

    obs = Observations({"smryh": [{"key": "FOPT", "histvec": "FOPTH"}]})

    mismatch = obs.mismatch(ens)

    assert "L1" in mismatch.columns
    assert "L2" in mismatch.columns
    assert "MISMATCH" in mismatch.columns
    assert "OBSKEY" in mismatch.columns
    assert "OBSTYPE" in mismatch.columns
    assert "REAL" in mismatch.columns
    assert len(mismatch) == len(ens) * 1  # number of observation units.

    fopt_rank = mismatch.sort_values("L2", ascending=True)["REAL"].values
    assert fopt_rank[0] == 2  # closest realization
    assert fopt_rank[-1] == 1  # worst realization

    # Try again with reference to non-existing vectors:
    obs = Observations({"smryh": [{"key": "FOPTFLUFF", "histvec": "FOPTFLUFFH"}]})
    mismatch = obs.mismatch(ens)
    assert mismatch.empty


def test_vens_mismatch():
    """Test calculation of mismatch to virtualized ensemble data"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")
    ens = ScratchEnsemble(
        "test", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/"
    )
    ens.load_smry(column_keys=["FOPT*"], time_index="monthly")

    vens = ens.to_virtual()

    # We don't need time_index now, because monthly is all we have.
    obs = Observations({"smryh": [{"key": "FOPT", "histvec": "FOPTH"}]})

    mismatch = obs.mismatch(vens)
    mismatch_raw = obs.mismatch(ens)
    assert isinstance(mismatch, pd.DataFrame)
    assert not mismatch.empty
    assert "L1" in mismatch.columns
    assert "L2" in mismatch.columns
    assert "MISMATCH" in mismatch.columns

    assert mismatch["MISMATCH"].sum() != mismatch_raw["MISMATCH"].sum()

    obs_monthly = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "monthly"}]}
    )
    assert (
        (
            mismatch.sort_values("REAL")
            .reset_index(drop=True)
            .drop("TIME_INDEX", axis=1)
            == obs_monthly.mismatch(ens)
            .sort_values("REAL")
            .reset_index(drop=True)
            .drop("TIME_INDEX", axis=1)
        )
        .all()
        .all()
    )

    # We should be able to do yearly smryh comparisons from virtualized
    # monthly profiles:
    obs_yearly = Observations(
        {"smryh": [{"key": "FOPT", "histvec": "FOPTH", "time_index": "yearly"}]}
    )
    mismatch_yearly = obs_yearly.mismatch(vens)
    assert mismatch_yearly["MISMATCH"].sum() != mismatch["MISMATCH"].sum()

    # When load_smry() is forgotten before virtualization:
    vens = ScratchEnsemble(
        "test", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/"
    ).to_virtual()
    with pytest.raises(ValueError):
        obs.mismatch(vens)

    # Removal of one realization in the virtualized ensemble:
    ens = ScratchEnsemble(
        "test", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/"
    )
    ens.load_smry(column_keys=["FOPT*"], time_index="monthly")
    vens = ens.to_virtual()
    vens.remove_realizations(2)
    mismatch_subset = obs.mismatch(vens)
    assert 2 not in mismatch_subset["REAL"].unique()
    assert 0 in mismatch_subset["REAL"].unique()


def test_ens_failedreals():
    """Ensure we can calculate mismatch where some realizations
    do not have UNSMRY data"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")
    ens = ScratchEnsemble(
        "test",
        testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/",
        autodiscovery=False,
    )
    obs = Observations({"smryh": [{"key": "FOPT", "histvec": "FOPTH"}]})
    mismatch = obs.mismatch(ens)

    # There are no UNSMRY found, so the mismatch should be empty:
    assert mismatch.empty

    ens.find_files("eclipse/model/*UNSMRY")
    assert not obs.mismatch(ens).empty

    # Reinitialize
    ens = ScratchEnsemble(
        "test",
        testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/",
        autodiscovery=False,
    )

    # Redirect UNSMRY pointer in realizaion 3 so it isn't found
    ens.find_files("eclipse/model/*UNSMRY")
    real3files = ens[3].files
    real3files.loc[real3files["FILETYPE"] == "UNSMRY", "FULLPATH"] = "FOO"

    # Check that we only have EclSum for 2 and not for 3:
    assert ens[2].get_eclsum()
    assert not ens[3].get_eclsum()

    missingsmry = obs.mismatch(ens)
    # Realization 3 should NOT be present now
    assert 3 not in list(missingsmry["REAL"])
    assert not obs.mismatch(ens).empty


def test_ensset_mismatch():
    """Test mismatch calculation on an EnsembleSet"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    ensdir = os.path.join(testdir, "data/testensemble-reek001/")

    # Copy iter-0 to iter-1, creating an identical ensemble
    # we can load for testing.
    for realizationdir in glob.glob(ensdir + "/realization-*"):
        if os.path.exists(realizationdir + "/iter-1"):
            os.remove(realizationdir + "/iter-1")
        os.symlink(realizationdir + "/iter-0", realizationdir + "/iter-1")

    iter0 = ScratchEnsemble("iter-0", ensdir + "/realization-*/iter-0")
    iter1 = ScratchEnsemble("iter-1", ensdir + "/realization-*/iter-1")

    ensset = EnsembleSet("reek001", [iter0, iter1])

    obs = Observations({"smryh": [{"key": "FOPT", "histvec": "FOPTH"}]})

    mismatch = obs.mismatch(ensset)
    assert "ENSEMBLE" in mismatch.columns
    assert "REAL" in mismatch.columns
    assert len(mismatch) == 10
    assert (
        mismatch[mismatch.ENSEMBLE == "iter-0"].L1.sum()
        == mismatch[mismatch.ENSEMBLE == "iter-1"].L1.sum()
    )

    # This is quite hard to input in dict-format. Better via YAML..
    obs_pr = Observations(
        {
            "smry": [
                {
                    "key": "WBP4:OP_1",
                    "comment": "Pressure observations well OP_1",
                    "observations": [
                        {"value": 250, "error": 1, "date": datetime.date(2001, 1, 1)}
                    ],
                }
            ]
        }
    )

    mis_pr = obs_pr.mismatch(ensset)
    assert len(mis_pr) == 10

    # We should also be able to input dates as strings, and they
    # should be attempted parsed to datetime.date:
    obs_pr = Observations(
        {
            "smry": [
                {
                    "key": "WBP4:OP_1",
                    "observations": [{"value": 250, "error": 1, "date": "2001-01-01"}],
                }
            ]
        }
    )
    mis_pr2 = obs_pr.mismatch(ensset)
    assert len(mis_pr2) == 10

    # We are strict and DO NOT ALLOW non-ISO dates like this:
    with pytest.raises(ValueError):
        obs_pr = Observations(
            {
                "smry": [
                    {
                        "key": "WBP4:OP_1",
                        "observations": [
                            {"value": 250, "error": 1, "date": "01-01-2001"}
                        ],
                    }
                ]
            }
        )

    # Erroneous date will raise Exception
    # (but a valid date will give an extrapolated value)
    with pytest.raises(ValueError):
        obs_pr = Observations(
            {
                "smry": [
                    {
                        "key": "WBP4:OP_1",
                        "observations": [
                            {"value": 250, "error": 1, "date": "3011-45-443"}
                        ],
                    }
                ]
            }
        )
    obs_extrap = Observations(
        {
            "smry": [
                {
                    "key": "WBP4:OP_1",
                    "observations": [{"value": 250, "error": 1, "date": "1977-01-01"}],
                }
            ]
        }
    )
    assert len(obs_extrap.mismatch(ensset)) == 10  # (5 reals, 2 ensembles)


def test_virtual_observations():
    """Construct an virtual(?) observation object from a specific summary vector
    and use it to rank realizations for similarity.
    """

    # We need an ensemble to work with:
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")
    ens = ScratchEnsemble(
        "test", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0/"
    )
    ens.load_smry(
        column_keys=["FOPT", "FGPT", "FWPT", "FWCT", "FGOR"], time_index="yearly"
    )

    # And we need some VirtualRealizations
    virtreals = {
        "p90realization": ens.agg("p90"),
        "meanrealization": ens.agg("mean"),
        "p10realization": ens.agg("p10"),
    }

    summaryvector = "FOPT"
    representative_realizations = {}
    for virtrealname, virtreal in virtreals.items():
        # Create empty observation object
        obs = Observations({})
        obs.load_smry(virtreal, summaryvector, time_index="yearly")

        # Calculate how far each realization is from this observation set
        # (only one row pr. realization, as FOPTH is only one observation unit)
        mis = obs.mismatch(ens)

        closest_realization = (
            mis.groupby("REAL")["L2"].sum().sort_values().index.values[0]
        )
        representative_realizations[virtrealname] = closest_realization

    assert representative_realizations["meanrealization"] == 4
    assert representative_realizations["p90realization"] == 2
    assert representative_realizations["p10realization"] == 1

    # Test again with the ensemble virtualized:
    vens = ens.to_virtual()

    # And we need some VirtualRealizations
    vvirtreals = {
        "p90realization": vens.agg("p90"),
        "meanrealization": vens.agg("mean"),
        "p10realization": vens.agg("p10"),
    }

    summaryvector = "FOPT"
    vrepresentative_realizations = {}
    for virtrealname, virtreal in vvirtreals.items():
        # Create empty observation object
        obs = Observations({})
        obs.load_smry(virtreal, summaryvector, time_index="yearly")

        mis = obs.mismatch(ens)

        closest_realization = (
            mis.groupby("REAL")["L2"].sum().sort_values().index.values[0]
        )
        vrepresentative_realizations[virtrealname] = closest_realization

    assert vrepresentative_realizations["meanrealization"] == 4
    assert vrepresentative_realizations["p90realization"] == 2
    assert vrepresentative_realizations["p10realization"] == 1



"""
Observations support and related calculations
"""

import datetime
import logging
import math
import os
from collections import OrderedDict

import dateutil
import pandas as pd
import yaml

from .ensemble import ScratchEnsemble
from .ensemblecombination import EnsembleCombination
from .ensembleset import EnsembleSet
from .realization import ScratchRealization
from .realizationcombination import RealizationCombination
from .virtualensemble import VirtualEnsemble
from .virtualrealization import VirtualRealization

logger = logging.getLogger(__name__)


class Observations(object):
    """Represents a set of observations and the ability to
    compare realizations and ensembles to the observations

    The primary data structure is a dictionary holding actual
    observations, this can typically be loaded from a yaml-file

    Key functionality is to be able to compute mismatch pr
    observation and presenting the computed data as a Pandas
    Dataframe. If run on ensembles, every row will be tagged
    by which realization index the data was computed for.

    An observation unit is a concept for the observation and points to
    something we define as a "single" observation. It can be one value
    for one datatype at a specific date, but in the case of Eclipse
    summary vector, it can also be a time-series. Mismatches will
    be computed pr. observation unit.

    Pay attention to mismatch versus misfit. Here, mismatch is used
    for individual observation units, while misfit is used as single
    number for whole realizations.

    Important: Using time-series as observations is not recommended in
    assisted history match. Pick individual uncorrelated data points
    at relevant points in time instead.

    The type of observations supported must follow the datatypes that
    the realizations and ensemble objects are able to internalize.

    """

    # Discussion points:
    # * Should mismatch calculation happen in this function
    #   with ensembles/realizations input or the other way around?
    # * Should it be possible to represent the observations
    #   themselves in a dataframe, or does the dict suffice?
    #   (each observation unit should be representable as
    #   a dict, and then it is mergeable in Pandas)

    def __init__(self, observations):
        """Initialize an observation object with observations
        from file or from an incoming dictionary structure

        Observations will be checked for validity, and
        incorrect observations (wrong format, unsupported etc.)
        will be removed. Empty observation list is allowed, and
        will typically end in empty result dataframes

        Args:
            observations: dict with observation structure or string
                with path to a yaml file.
        """
        self.observations = dict()

        if isinstance(observations, str):
            with open(observations) as yamlfile:
                self.observations = yaml.full_load(yamlfile)
        elif isinstance(observations, dict):
            self.observations = observations
        else:
            raise ValueError("Unsupported object for observations")

        # Remove unsupported observations
        # Identify and warn about errors in observation syntax (dates etc)
        self._clean_observations()

        logger.info("Initialized observation with obstypes %s", str(self.keys()))
        for obskey in self.keys():
            # (fixme: this string does not make sense)
            logger.info(" %s: ", str(len(self.observations[obskey])))

    def __getitem__(self, someobject):
        """Pick objects from the observations dict"""
        return self.observations[someobject]

    def mismatch(self, ens_or_real):
        """Compute the mismatch from the current observation set
        to the incoming ensemble or realization.

        In the case of an ensemble, it will calculate individually
        for every realization, and aggregate the results.

        Returns:
            dataframe with REAL (only if ensemble), OBSKEY, DATE,
                L1, L2. One row for every observation unit.
        """
        # For ensembles, we should in the future be able to loop
        # over realizations in a multiprocessing fashion
        if isinstance(ens_or_real, EnsembleCombination):
            logger.info("Evaluating EnsembleCombination")
            ens_or_real = ens_or_real.to_virtual()
        if isinstance(ens_or_real, RealizationCombination):
            logger.info("Evaluating RealizationCombination")
            ens_or_real = ens_or_real.to_virtual()
        if isinstance(ens_or_real, EnsembleSet):
            mismatches = {}
            # pylint: disable=protected-access
            for ensname, ens in ens_or_real._ensembles.items():
                logger.info("Calculating mismatch for ensemble %s", ensname)
                for realidx, real in ens.realizations.items():
                    logger.info("Calculating mismatch for realization %s", str(realidx))
                    mismatches[(ensname, realidx)] = self._realization_mismatch(real)
                    mismatches[(ensname, realidx)]["REAL"] = realidx
                    mismatches[(ensname, realidx)]["ENSEMBLE"] = ensname
            return pd.concat(mismatches, axis=0, ignore_index=True)
        if isinstance(ens_or_real, ScratchEnsemble):
            mismatches = {}
            for realidx, real in ens_or_real.realizations.items():
                mismatches[realidx] = self._realization_mismatch(real)
                mismatches[realidx]["REAL"] = realidx
            return pd.concat(mismatches, axis=0, ignore_index=True, sort=False)
        if isinstance(ens_or_real, VirtualEnsemble):
            logger.info("Calculating mismatch on ensemble %s", ens_or_real.name)
            mismatches = {}
            for realidx in ens_or_real.realindices:
                mismatches[realidx] = self._realization_mismatch(
                    ens_or_real.get_realization(realidx)
                )
                mismatches[realidx]["REAL"] = realidx
            return pd.concat(mismatches, axis=0, ignore_index=True, sort=False)
        if isinstance(ens_or_real, (ScratchRealization, VirtualRealization)):
            return self._realization_mismatch(ens_or_real)
        if isinstance(ens_or_real, EnsembleSet):
            raise NotImplementedError
        raise ValueError("Unsupported object for mismatch calculation")

    def load_smry(self, realization, smryvector, time_index="yearly", smryerror=None):
        """Add an observation unit from a VirtualRealization or
        ScratchRealization, being a specific summaryvector, picking
        values with the specified time resolution.

        This can be used to compare similarity between realization, by
        viewing simulated results as "observations". A use case is
        to rank all realizations in an ensemble for the similarity to
        a certain mean profile, f.ex. FOPT.

        The result of the function is a observation unit added to
        the smry observations, with values at every date.

        Arguments:
            realization: ScratchRealization or VirtualRealization containing
                data for constructing the virtual observation
            smryvector: string with a name of a specific summary vector
                to be used
            time_index: string with timeresolution, typically 'yearly'
                or 'monthly'.
            smryerror: float, constant value to be used as the measurement
                error for every date.
        """
        dataseries = realization.get_smry(
            column_keys=[smryvector], time_index=time_index
        )[smryvector]

        # In the context of this function, datetimes are not supported. Ensure dates:
        if isinstance(dataseries.index, pd.DatetimeIndex):
            dataseries.index = dataseries.index.date
        # If the index is a list of strings (object),
        # it is ok (assuming ISO-datestrings)

        # Modify the observation object (self)
        if "smry" not in self.observations.keys():
            self.observations["smry"] = []  # Empty list

        # Construct a virtual observation with observation units
        # at every timestep:
        virtobs = {}
        virtobs["key"] = smryvector
        virtobs["comment"] = "Virtual observation unit constructed from " + str(
            realization
        )
        virtobs["observations"] = []
        for date, value in dataseries.items():
            virtobs["observations"].append(
                {"value": value, "error": smryerror, "date": date}
            )
        self.observations["smry"].append(virtobs)

    def __len__(self):
        """Return the number of observation units present"""
        # This is not correctly implemented yet..
        return len(self.observations.keys())

    @property
    def empty(self):
        """Decide if the observation set is empty

        An empty observation set is has zero observation
        unit count"""
        return not self.__len__()

    def keys(self):
        """Return a list of observation units present.

        This list might change into a dataframe in the future,
        but calling len() on its results should always return
        the number of observation units."""
        return self.observations.keys()

    def _realization_mismatch(self, real):
        """Compute the mismatch from the current loaded
        observations to a realization.

        Supports both ScratchRealizations and
        VirtualRealizations

        The returned dataframe contains the columns:
            * OBSTYPE - category/type of the observation
            * OBSKEY - name of the observation key
            * LABEL - if an observation has it
            * DATE - only where relevant.
            * OBSINDEX - where an enumeration is relevant
            * MISMATCH - signed difference between value and result
            * L1 - absolute difference
            * L2 - absolute difference squared
            * SIGN - True if positive difference
            * SIMVALUE - the simulated value, not for smryh
            * OBSVALUE - the observed value, not for smryh
        One row for every observation unit.

        Args:
            real : ScratchRealization or VirtualRealization
        Returns:
            dataframe: One row per observation unit with
                mismatch data
        """
        # mismatch_df = pd.DataFrame(columns=['OBSTYPE', 'OBSKEY',
        #     'DATE', 'OBSINDEX', 'MISMATCH', 'L1', 'L2', 'SIGN'])
        mismatches = []
        for obstype in self.observations.keys():
            for obsunit in self.observations[obstype]:  # (list)
                if obstype == "txt":
                    try:
                        sim_value = real.get_df(obsunit["localpath"])[obsunit["key"]]
                    except (KeyError, ValueError):
                        logger.warning(
                            "%s in %s not found, ignored",
                            obsunit["key"],
                            obsunit["localpath"],
                        )
                        continue
                    mismatch = float(sim_value - obsunit["value"])
                    measerror = 1
                    sign = (mismatch > 0) - (mismatch < 0)
                    mismatches.append(
                        dict(
                            OBSTYPE=obstype,
                            OBSKEY=str(obsunit["localpath"])
                            + "/"
                            + str(obsunit["key"]),
                            LABEL=obsunit.get("label", ""),
                            MISMATCH=mismatch,
                            L1=abs(mismatch),
                            L2=abs(mismatch) ** 2,
                            SIMVALUE=sim_value,
                            OBSVALUE=obsunit["value"],
                            MEASERROR=measerror,
                            SIGN=sign,
                        )
                    )
                if obstype == "scalar":
                    try:
                        sim_value = real.get_df(obsunit["key"])
                    except (KeyError, ValueError):
                        logger.warning(
                            "No data found for scalar: %s, ignored", obsunit["key"]
                        )
                        continue
                    mismatch = float(sim_value - obsunit["value"])
                    measerror = 1
                    sign = (mismatch > 0) - (mismatch < 0)
                    mismatches.append(
                        dict(
                            OBSTYPE=obstype,
                            OBSKEY=str(obsunit["key"]),
                            LABEL=obsunit.get("label", ""),
                            MISMATCH=mismatch,
                            L1=abs(mismatch),
                            SIMVALUE=sim_value,
                            OBSVALUE=obsunit["value"],
                            MEASERROR=measerror,
                            L2=abs(mismatch) ** 2,
                            SIGN=sign,
                        )
                    )
                if obstype == "smryh":
                    if "time_index" in obsunit:
                        if isinstance(obsunit["time_index"], str):
                            sim_hist = real.get_smry(
                                time_index=obsunit["time_index"],
                                column_keys=[obsunit["key"], obsunit["histvec"]],
                            )
                        elif isinstance(
                            obsunit["time_index"], (datetime.datetime, datetime.date)
                        ):
                            # real.get_smry only allows strings or
                            # list of datetimes as time_index.
                            sim_hist = real.get_smry(
                                time_index=[obsunit["time_index"]],
                                column_keys=[obsunit["key"], obsunit["histvec"]],
                            )
                        else:
                            logger.error(
                                (
                                    "obsunit-timeindex was not string or date object\n"
                                    "Should not be possible, file a bug report"
                                )
                            )
                            logger.error(obsunit["time_index"])
                            logger.error(type(obsunit["time_index"]))
                        time_index_str = str(obsunit["time_index"])
                    else:
                        sim_hist = real.get_smry(
                            column_keys=[obsunit["key"], obsunit["histvec"]]
                            # (let get_smry() determine the possible time_index)
                        )
                        time_index_str = ""
                    # If empty df returned, we don't have the data for this:
                    if sim_hist.empty:
                        logger.warning(
                            "No data found for smryh: %s and %s, ignored.",
                            obsunit["key"],
                            obsunit["histvec"],
                        )
                        continue
                    sim_hist["mismatch"] = (
                        sim_hist[obsunit["key"]] - sim_hist[obsunit["histvec"]]
                    )
                    measerror = 1
                    mismatches.append(
                        dict(
                            OBSTYPE="smryh",
                            OBSKEY=obsunit["key"],
                            LABEL=obsunit.get("label", ""),
                            MISMATCH=sim_hist["mismatch"].sum(),
                            MEASERROR=measerror,
                            L1=sim_hist["mismatch"].abs().sum(),
                            L2=math.sqrt((sim_hist["mismatch"] ** 2).sum()),
                            TIME_INDEX=time_index_str,
                        )
                    )
                if obstype == "smry":
                    # For 'smry', there is a list of
                    # observations (indexed by date)
                    for unit in obsunit["observations"]:
                        try:
                            sim_value = real.get_smry(
                                time_index=[unit["date"]], column_keys=obsunit["key"]
                            )[obsunit["key"]].values[0]
                        except KeyError:
                            logger.warning(
                                "No data found for smry: %s at %s, ignored.",
                                obsunit["key"],
                                str(unit["date"]),
                            )
                            continue
                        mismatch = float(sim_value - unit["value"])
                        sign = (mismatch > 0) - (mismatch < 0)
                        mismatches.append(
                            dict(
                                OBSTYPE="smry",
                                OBSKEY=obsunit["key"],
                                DATE=unit["date"],
                                MEASERROR=unit["error"],
                                LABEL=unit.get("label", ""),
                                MISMATCH=mismatch,
                                OBSVALUE=unit["value"],
                                SIMVALUE=sim_value,
                                L1=abs(mismatch),
                                L2=abs(mismatch) ** 2,
                                SIGN=sign,
                            )
                        )
        return pd.DataFrame(mismatches)

    def _realization_misfit(self, real, defaulterrors=False, corr=None):
        """The misfit value for the observation set

        Ref: https://wiki.statoil.no/wiki/index.php/RP_HM/Observations#Misfit_function

        Args:
            real : a ScratchRealization or a VirtualRealization
            defaulterrors: (boolean) If set to True, zero measurement errors
                will be set to 1.
            corr : correlation or weigthing matrix (numpy matrix).
                If a list or numpy vector is supplied, it is interpreted
                as a diagonal matrix. If omitted, the identity matrix is used

        Returns:
            float : the misfit value for the observation set and realization
        """  # noqa
        if corr:
            raise NotImplementedError(
                "correlations in misfit " + "calculation is not supported"
            )
        mismatch = self._realization_mismatch(real)

        zeroerrors = mismatch["MEASERROR"] < 1e-7
        if defaulterrors:
            mismatch[zeroerrors]["MEASERROR"] = 1
        else:
            if zeroerrors.any():
                print(mismatch[zeroerrors])
                raise ValueError(
                    "Zero measurement error in observation set"
                    + ". can't be used to calculate misfit"
                )
        if "MISFIT" not in mismatch.columns:
            mismatch["MISFIT"] = mismatch["L2"] / (mismatch["MEASERROR"] ** 2)

        return mismatch["MISFIT"].sum()

    def _clean_observations(self):
        """Verify integrity of observations, remove
        observation units that cannot be used.

        Will log warnings about things that are removed.

        Returns number of usable observation units.

        Ensure that dates are parsed into datetime.date objects.
        """
        supported_categories = ["smry", "smryh", "txt", "scalar", "rft"]

        # Check top level keys in observations dict:
        for key in list(self.observations):
            if key not in supported_categories:
                self.observations.pop(key)
                logger.error("Observation category %s not supported", key)
                continue
            if not isinstance(self.observations[key], list):
                logger.error(
                    "Observation category %s did not contain a list, but %s",
                    key,
                    type(self.observations[key]),
                )
                self.observations.pop(key)
        # Check smryh observations for validity
        if "smryh" in self.observations.keys():
            smryhunits = self.observations["smryh"]
            if not isinstance(smryhunits, list):
                logger.warning(
                    "smryh must consist of a list, deleting: %s", str(smryhunits)
                )
                del self.observations["smryh"]
            for unit in smryhunits:
                if not isinstance(unit, (dict, OrderedDict)):
                    logger.warning("smryh-units must be dicts, deleting: %s", str(unit))
                    del smryhunits[smryhunits.index(unit)]
                    continue
                if not ("key" in unit and "histvec" in unit):
                    logger.warning(
                        (
                            "smryh units must contain both 'key' and "
                            "'histvec', deleting: %s",
                            str(unit),
                        )
                    )
                    del smryhunits[smryhunits.index(unit)]
                    continue
                # If time_index is not a supported mnemonic,
                # parse it to a date object
                if "time_index" in unit:
                    if unit["time_index"] not in [
                        "raw",
                        "report",
                        "yearly",
                        "daily",
                        "first",
                        "last",
                        "monthly",
                    ] and not isinstance(unit["time_index"], datetime.datetime):
                        try:
                            unit["time_index"] = dateutil.parser.isoparse(
                                unit["time_index"]
                            ).date()
                        except (TypeError, ValueError) as exception:
                            logger.warning(
                                "Parsing date %s failed with error",
                                (str(unit["time_index"]), str(exception)),
                            )
                            del smryhunits[smryhunits.index(unit)]
                            continue
            # If everything has been deleted through cleanup, delete the section
            if not smryhunits:
                del self.observations["smryh"]
        # Check smry observations for validity
        if "smry" in self.observations.keys():
            # We already know that observations['smry'] is a list
            # Each list element must be a dict with
            # the mandatory keys 'key' and 'observation'
            smryunits = self.observations["smry"]
            for unit in smryunits:
                if not isinstance(unit, (dict, OrderedDict)):
                    logger.warning(
                        "Observation units must be dicts, deleting: %s", str(unit)
                    )
                    del smryunits[smryunits.index(unit)]
                    continue
                if not ("key" in unit and "observations" in unit):
                    logger.warning(
                        (
                            "Observation unit must contain key and",
                            "observations, deleting: %s",
                        ),
                        str(unit),
                    )
                    del smryunits[smryunits.index(unit)]
                    continue
                # Check if strings need to be parsed as dates:
                for observation in unit["observations"]:
                    if isinstance(observation["date"], str):
                        observation["date"] = dateutil.parser.isoparse(
                            observation["date"]
                        ).date()
                    if not isinstance(observation["date"], datetime.date):
                        logger.error("Date not understood %s", str(observation["date"]))
                        continue
            # If everything is deleted from 'smry', delete it
            if not smryunits:
                del self.observations["smry"]

    def to_ert2observations(self):
        """Convert the observation set to an observation
        file for use with Ert 2.x.

        Returns: multiline string
        """
        raise NotImplementedError

    def __repr__(self):
        """Return a representation of the object

        The representation is a YAML string
        that can be used to reinstatiate the
        object
        """
        return self.to_yaml()

    def to_yaml(self):
        """Convert the current observations to YAML format

        Returns:
            string : Multiline YAML string.
        """
        return yaml.dump(self.observations)

    def to_disk(self, filename):
        """Write the current observation object to disk

        In YAML-format. If a new observation object
        is instantiated from the outputted filename, it
        should yield identical results in mismatch
        calculation.

        Directory structure will be created if not existing.
        Existing file will be overwritten.

        Arguments:
            filename - string with path and filename to
                be written to"""
        if not isinstance(filename, str):
            raise ValueError("Filename must be a string")
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname) and dirname:
            os.makedirs(dirname)
        with open(filename, "w") as fhandle:
            fhandle.write(self.to_yaml())



"""Common utility functions used in fmu.ensemble"""

import datetime
import logging
from typing import List, Tuple

import dateutil
import pandas as pd

logger = logging.getLogger(__name__)

"""Mapping from fmu-ensemble custom offset strings to Pandas DateOffset strings.
See
https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
"""
PD_FREQ_MNEMONICS = {
    "monthly": "MS",
    "yearly": "YS",
    "daily": "D",
    "weekly": "W-MON",
}


def date_range(start_date, end_date, freq):
    """Wrapper for pandas.date_range to allow for extra fmu-ensemble specific mnemonics
    'yearly', 'daily', 'weekly', mapped over to pandas DateOffsets

    Args:
        start_date (datetime.date)
        end_date (datetime.date)
        freq (str): monthly, daily, weekly, yearly, or a Pandas date offset
            frequency.

    Returns:
        list of datetimes
    """
    try:
        return pd.date_range(
            start_date, end_date, freq=PD_FREQ_MNEMONICS.get(freq, freq)
        )
    except pd.errors.OutOfBoundsDatetime:
        return _fallback_date_range(start_date, end_date, freq)


def unionize_smry_dates(eclsumsdates, freq, normalize, start_date=None, end_date=None):
    """
    Unionize lists of dates into one datelist encompassing the date
    range from all datelists, with cropping, resampling to wanted
    frequency, and potential normalized dates to the frequency (rollback/rollforward)

    Args:
        eclsumsdates (list of lists of datetimes)
        freq (str): Requested frequency
        normalize (bool): Normalize daterange to frequency or not.
        start_date (datetime.date or str): Overridden if freq=='first'
        end_date (datetime.date or str): Overridden if freq=='last'

    Return:
        list of datetime.date
    """
    if not eclsumsdates:
        return []

    if start_date and isinstance(start_date, str):
        start_date = dateutil.parser.isoparse(start_date).date()
    if start_date and not isinstance(start_date, datetime.date):
        raise TypeError("start_date had unknown type")

    if end_date and isinstance(end_date, str):
        end_date = dateutil.parser.isoparse(end_date).date()
    if end_date and not isinstance(end_date, datetime.date):
        raise TypeError("end_date had unknown type")

    if freq in ("report", "raw"):
        datetimes = set()
        for eclsumdatelist in eclsumsdates:
            datetimes = datetimes.union(eclsumdatelist)
        datetimes = list(datetimes)
        datetimes.sort()
        if start_date:
            # Convert to datetime (at 00:00:00)
            start_date = datetime.datetime.combine(
                start_date, datetime.datetime.min.time()
            )
            datetimes = [x for x in datetimes if x > start_date]
            datetimes = [start_date] + datetimes
        if end_date:
            end_date = datetime.datetime.combine(end_date, datetime.datetime.min.time())
            datetimes = [x for x in datetimes if x < end_date]
            datetimes = datetimes + [end_date]
        return datetimes
    if freq == "last":
        end_date = max([max(x) for x in eclsumsdates]).date()
        return [end_date]
    if freq == "first":
        start_date = min([min(x) for x in eclsumsdates]).date()
        return [start_date]
    # These are datetime.datetime, not datetime.date
    start_smry = min([min(x) for x in eclsumsdates])
    end_smry = max([max(x) for x in eclsumsdates])

    (start_n, end_n) = normalize_dates(start_smry.date(), end_smry.date(), freq)

    if not start_date and not normalize:
        start_date_range = start_smry.date()
    elif not start_date and normalize:
        start_date_range = start_n
    else:
        start_date_range = start_date

    if not end_date and not normalize:
        end_date_range = end_smry.date()
    elif not end_date and normalize:
        end_date_range = end_n
    else:
        end_date_range = end_date

    datetimes = date_range(start_date_range, end_date_range, freq)

    # Convert from Pandas' datetime64 to datetime.date:
    datetimes = [x.date() for x in datetimes]

    # pd.date_range will not include random dates that do not
    # fit on frequency boundary. Force include these if
    # supplied as user arguments.
    if start_date and start_date not in datetimes:
        datetimes = [start_date] + datetimes
    if end_date and end_date not in datetimes:
        datetimes = datetimes + [end_date]
    return datetimes


def normalize_dates(
    start_date: datetime.date, end_date: datetime.date, freq: str
) -> Tuple[datetime.date, datetime.date]:
    """
    Normalize start and end date according to frequency
    by extending the time range.

    So for [1997-11-5, 2020-03-02] and monthly freqency
    this will transform your dates to
    [1997-11-1, 2020-04-01]

    For yearly frequency will be [1997-01-01, 2021-01-01].

    Args:
        start_date: datetime.date
        end_date: datetime.date
        freq: string with either 'monthly', 'yearly', 'weekly'
            or any other frequency offset accepted by Pandas

    Return:
        Tuple of normalized (start_date, end_date)
    """
    offset = pd.tseries.frequencies.to_offset(PD_FREQ_MNEMONICS.get(freq, freq))
    try:
        start_normalized = offset.rollback(start_date).date()
    except pd.errors.OutOfBoundsDatetime:
        # Pandas only supports datetime up to year 2262
        start_normalized = _fallback_date_roll(
            datetime.datetime.combine(start_date, datetime.time()), "back", freq
        ).date()
    try:
        end_normalized = offset.rollforward(end_date).date()
    except pd.errors.OutOfBoundsDatetime:
        # Pandas only supports datetime up to year 2262
        end_normalized = _fallback_date_roll(
            datetime.datetime.combine(end_date, datetime.time()), "forward", freq
        ).date()

    return (start_normalized, end_normalized)


def _fallback_date_roll(
    rollme: datetime.datetime, direction: str, freq: str
) -> datetime.datetime:
    """Fallback function for rolling dates forward or backward onto a
    date frequency boundary.

    This function reimplements pandas' DateOffset.roll_forward() and backward()
    only for monthly and yearly frequency. This is necessary as Pandas does not
    support datetimes beyond year 2262 due to all datetimes in Pandas being
    represented by nanosecond accuracy.

    This function is a fallback only, to keep support for using all Pandas timeoffsets
    in situations where years beyond 2262 is not a issue."""
    if direction not in ["back", "forward"]:
        raise ValueError(f"Unknown direction {direction}")

    if freq == "yearly":
        if direction == "forward":
            if rollme <= datetime.datetime(year=rollme.year, month=1, day=1):
                return datetime.datetime(year=rollme.year, month=1, day=1)
            return datetime.datetime(year=rollme.year + 1, month=1, day=1)
        return datetime.datetime(year=rollme.year, month=1, day=1)

    if freq == "monthly":
        if direction == "forward":
            if rollme <= datetime.datetime(year=rollme.year, month=rollme.month, day=1):
                return datetime.datetime(year=rollme.year, month=rollme.month, day=1)
            return datetime.datetime(
                year=rollme.year, month=rollme.month, day=1
            ) + dateutil.relativedelta.relativedelta(  # type: ignore
                months=1
            )
        return datetime.datetime(year=rollme.year, month=rollme.month, day=1)

    raise ValueError(
        "Only yearly or monthly frequencies are "
        "supported for simulations beyond year 2262"
    )


def _fallback_date_range(
    start: datetime.date, end: datetime.date, freq: str
) -> List[datetime.datetime]:
    """Fallback routine for generating date ranges beyond Pandas datetime64[ns]
    year-2262 limit.

    Assumes that the start and end times already fall on a frequency boundary.
    """
    if start == end:
        return [datetime.datetime.combine(start, datetime.datetime.min.time())]
    if end < start:
        return []
    if freq == "yearly":
        dates = [datetime.datetime.combine(start, datetime.datetime.min.time())] + [
            datetime.datetime(year=year, month=1, day=1)
            for year in range(start.year + 1, end.year + 1)
        ]
        if datetime.datetime.combine(end, datetime.datetime.min.time()) != dates[-1]:
            dates = dates + [
                datetime.datetime.combine(end, datetime.datetime.min.time())
            ]
        return dates
    if freq == "monthly":
        dates = []
        date = datetime.datetime.combine(start, datetime.datetime.min.time())
        enddatetime = datetime.datetime.combine(end, datetime.datetime.min.time())
        while date <= enddatetime:
            dates.append(date)
            date = date + dateutil.relativedelta.relativedelta(months=1)  # type: ignore
        return dates
    raise ValueError("Unsupported frequency for datetimes beyond year 2262")



"""Testing incorporation of ecl2df in fmu-ensemble."""

import logging
import os

import pytest
from fmu.ensemble import ScratchEnsemble, ScratchRealization

HAVE_ECL2DF = True
try:
    import ecl2df
except ImportError:
    HAVE_ECL2DF = False

logger = logging.getLogger(__name__)


def test_ecl2df_real():
    """Check that we can utilize ecl2df on single realizations"""

    if not HAVE_ECL2DF:
        pytest.skip()

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")
    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ScratchRealization(realdir)

    eclfiles = real.get_eclfiles()
    assert isinstance(eclfiles, ecl2df.EclFiles)
    compdat_df = ecl2df.compdat.df(eclfiles)
    assert not compdat_df.empty
    assert "KH" in compdat_df


def test_reek():
    """Import the reek ensemble and apply ecl2df functions on
    the realizations"""

    if "__file__" in globals():
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    reekens = ScratchEnsemble(
        "reektest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    if not HAVE_ECL2DF:
        pytest.skip()

    def extract_compdat(kwargs):
        """Callback fnction to extract compdata data using ecl2df
        on a ScratchRealization"""
        eclfiles = kwargs["realization"].get_eclfiles()
        if not eclfiles:
            print(
                "Could not obtain EclFiles object for realization "
                + str(kwargs["realization"].index)
            )
        return ecl2df.compdat.deck2dfs(eclfiles.get_ecldeck())["COMPDAT"]

    allcompdats = reekens.apply(extract_compdat)
    assert not allcompdats.empty
    assert 0 in allcompdats["REAL"]
    assert "KH" in allcompdats
    # Pr. now, only realization-0 has eclipse/include in git


def test_smry_via_ecl2df():
    """Test that we could use ecl2df for smry extraction instead
    of the native code inside fmu-ensemble"""

    def get_smry(kwargs):
        """Callback function to extract smry data using ecl2df on a
        ScratchRealization"""
        eclfiles = kwargs["realization"].get_eclfiles()
        return ecl2df.summary.df(
            eclfiles, time_index=kwargs["time_index"], column_keys=kwargs["column_keys"]
        )

    if "__file__" in globals():
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    reekens = ScratchEnsemble(
        "reektest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    if not HAVE_ECL2DF:
        pytest.skip()

    callback_smry = reekens.apply(get_smry, column_keys="FOPT", time_index="yearly")
    direct_smry = reekens.get_smry(column_keys="FOPT", time_index="yearly")

    assert callback_smry["FOPT"].sum() == direct_smry["FOPT"].sum()
    assert callback_smry["REAL"].sum() == direct_smry["REAL"].sum()
    # BUG in ecl2df, dates are missing!!



"""Module for the ScratchRealization class

A realization is a set of results from one subsurface model
realization. A realization can be either defined from
its output files from the FMU run on the file system,
it can be computed from other realizations, or it can be
an archived realization.
"""

import copy
import glob
import json
import logging
import os
import re
import warnings
from datetime import date, datetime, time

import dateutil
import numpy as np
import pandas as pd
import yaml
from resdata.grid import Grid
from resdata.rd_util import FileMode
from resdata.resfile import ResdataFile
from resdata.summary import Summary

from .realizationcombination import RealizationCombination
from .util import flatten, parse_number, shortcut2path
from .util.dates import unionize_smry_dates
from .util.rates import compute_volumetric_rates
from .virtualrealization import VirtualRealization

HAVE_ECL2DF = False
try:
    import ecl2df

    HAVE_ECL2DF = True
except ImportError:
    HAVE_ECL2DF = False

logger = logging.getLogger(__name__)


class ScratchRealization(object):
    r"""A representation of results still present on disk

    ScratchRealizations point to the filesystem for their
    contents.

    A realization must at least contain a STATUS file.
    Additionally, jobs.json and parameters.txt will be attempted
    loaded by default.

    The realization is defined by the pointers to the filesystem.
    When asked for, this object will return data from the
    filesystem (or from cache if already computed).

    The files dataframe is the central filesystem pointer
    repository for the object. It will at least contain
    the columns
    * FULLPATH absolute path to a file
    * FILETYPE filename extension (after last dot)
    * LOCALPATH relative filename inside realization diretory
    * BASENAME filename only. No path. Includes extension

    This dataframe is available as a read-only property from the object

    Args:
        path (str): absolute or relative path to a directory
            containing a realizations files.
        realidxregexp (re/str): a compiled regular expression which
            is used to determine the realization index (integer)
            from the path. First match is the index.
            Default: realization-(\d+)
            Only needs to match path components.
            If a string is supplied, it will be attempted
            compiled into a regular expression.
        index (int): the realization index to be used, will
            override anything else.
        autodiscovery (boolean): whether the realization should try to
            auto-discover certain data (UNSMRY files in standard location)
        batch (dict): List of functions (load_*) that
            should be run at time of initialization. Each element is a
            length 1 dictionary with the function name to run as the key
            and each keys value should be the function arguments as a dict.
    """

    def __init__(
        self, path, realidxregexp=None, index=None, autodiscovery=True, batch=None
    ):
        self._origpath = os.path.abspath(path)
        self.index = None
        self._autodiscovery = autodiscovery

        if not realidxregexp:
            realidxregexp = re.compile(r"realization-(\d+)")
        # Try to compile the regexp on behalf of the user.
        if isinstance(realidxregexp, str):
            realidxregexp = re.compile(realidxregexp)
        if isinstance(realidxregexp, str):
            raise ValueError("Supplied realidxregexp not valid")

        self.files = pd.DataFrame(
            columns=["FULLPATH", "FILETYPE", "LOCALPATH", "BASENAME"]
        )
        self._eclsum = None  # Placeholder for caching
        self._eclsum_include_restart = None  # Flag for cached object

        # The datastore for internalized data. Dictionary
        # indexed by filenames (local to the realization).
        # values in the dictionary can be either dicts or dataframes
        self.data = {}
        self._eclinit = None
        self._eclunrst = None
        self._eclgrid = None
        self._ecldata = None
        self._actnum = None

        abspath = os.path.abspath(path)

        if index is None:
            for path_comp in reversed(os.path.abspath(path).split(os.path.sep)):
                realidxmatch = re.match(realidxregexp, path_comp)
                if realidxmatch:
                    self.index = int(realidxmatch.group(1))
                    break
            else:
                logger.warning(
                    (
                        "Could not determine realization "
                        "index for %s, "
                        "this cannot be inserted in an Ensemble"
                    ),
                    abspath,
                )
                logger.warning("Maybe you need to use index=<someinteger>")
                self.index = None
        else:
            self.index = int(index)

        # Now look for some common files, but don't require any
        if os.path.exists(os.path.join(abspath, "STATUS")):
            filerow = {
                "LOCALPATH": "STATUS",
                "FILETYPE": "STATUS",
                "FULLPATH": os.path.join(abspath, "STATUS"),
                "BASENAME": "STATUS",
            }
            self.files = pd.concat(
                [self.files, pd.DataFrame([filerow])], ignore_index=True
            )
            self.load_status()
        else:
            logger.warning("No STATUS file, %s", abspath)

        if os.path.exists(os.path.join(abspath, "jobs.json")):
            filerow = {
                "LOCALPATH": "jobs.json",
                "FILETYPE": "json",
                "FULLPATH": os.path.join(abspath, "jobs.json"),
                "BASENAME": "jobs.json",
            }
            self.files = pd.concat(
                [self.files, pd.DataFrame([filerow])], ignore_index=True
            )

        if os.path.exists(os.path.join(abspath, "OK")):
            self.load_scalar("OK")

        if os.path.exists(os.path.join(abspath, "parameters.txt")):
            self.load_txt("parameters.txt")

        if batch:
            self.process_batch(batch)

        logger.info("Initialized %s", abspath)

    def process_batch(self, batch):
        """Process a list of functions to run/apply

        This is equivalent to calling each function individually
        but this enables more efficient concurrency. It is meant
        to be used for functions that modifies the realization
        object, not for functions that returns a dataframe already.

        Args:
            batch (list): Each list element is a dictionary with one key,
                being a function names, value pr key is a dict with keyword
                arguments to be supplied to each function.
        Returns:
            ScratchRealization: This realization object (self), for it
                to be picked up by ProcessPoolExecutor and pickling.
        """
        assert isinstance(batch, list)
        allowed_functions = [
            "apply",
            "find_files",
            "load_smry",
            "load_txt",
            "load_file",
            "load_csv",
            "load_status",
            "load_scalar",
        ]
        for cmd in batch:
            assert isinstance(cmd, dict)
            assert len(cmd) == 1
            fn_name = list(cmd.keys())[0]
            logger.info(
                "Batch processing (#%d): %s with args %s",
                self.index,
                fn_name,
                str(cmd[fn_name]),
            )
            if fn_name not in allowed_functions:
                logger.warning("process_batch skips illegal function: %s", fn_name)
                continue
            assert isinstance(cmd[fn_name], dict)
            getattr(self, fn_name)(**cmd[fn_name])
        return self

    def runpath(self):
        """Return the runpath ("root") of the realization

        Returns:
            str: the filesystem path which at least existed
                at time of object initialization.
        """
        return self._origpath

    def to_virtual(self, name=None, deepcopy=True):
        """Convert the current ScratchRealization object
        to a VirtualRealization

        Args:
            description (str): used as label
            deepcopy (boolean): Set to true if you want to continue
               to manipulate the ScratchRealization object
               afterwards without affecting the virtual realization.
               Defaults to True. False will give faster execution.
        """
        if not name:
            name = self._origpath
        if deepcopy:
            vreal = VirtualRealization(name, copy.deepcopy(self.data))
        else:
            vreal = VirtualRealization(name, self.data)

        # Conserve metadata for smry vectors. Build metadata dict for all
        # loaded summary vectors.
        smrycolumns = [
            self.get_df(key).columns for key in self.keys() if "unsmry" in key
        ]
        smrycolumns = {smrykey for sublist in smrycolumns for smrykey in sublist}
        meta = self.get_smry_meta(list(smrycolumns))
        if meta:
            meta_df = pd.DataFrame.from_dict(meta, orient="index")
            meta_df.index.name = "SMRYCOLUMN"
            vreal.append("__smry_metadata", meta_df.reset_index())
        return vreal

    def load_file(self, localpath, fformat, convert_numeric=True, force_reread=False):
        """
        Parse and internalize files from disk.

        Several file formats are supported:
        - txt (one key-value pair pr. line)
        - csv
        - scalar (one number or one string in the first line)
        """
        if fformat == "txt":
            self.load_txt(localpath, convert_numeric, force_reread)
        elif fformat == "csv":
            self.load_csv(localpath, convert_numeric, force_reread)
        elif fformat == "scalar":
            self.load_scalar(localpath, convert_numeric, force_reread)
        else:
            raise ValueError("Unsupported file format %s" % fformat)

    def load_scalar(
        self,
        localpath,
        convert_numeric=False,
        force_reread=False,
        comment=None,
        skip_blank_lines=True,
        skipinitialspace=True,
    ):
        """Parse a single value from a file.

        The value can be a string or a number.

        Empty files are treated as existing, with an empty string as
        the value, different from non-existing files.

        pandas.read_table() is used to parse the contents, the args
        'comment', 'skip_blank_lines', and 'skipinitialspace' is passed on
        to that function.

        Args:
            localpath: path to the file, local to the realization
            convert_numeric: If True, non-numerical content will be thrown away
            force_reread: Reread the data from disk.
        Returns:
            str/number: the value read from the file.
        """
        fullpath = os.path.abspath(os.path.join(self._origpath, localpath))
        if not os.path.exists(fullpath):
            raise IOError("File not found: " + fullpath)
        if fullpath in self.files["FULLPATH"].values and not force_reread:
            # Return cached version
            return self.data[localpath]
        if fullpath not in self.files["FULLPATH"].values:
            filerow = {
                "LOCALPATH": localpath,
                "FILETYPE": localpath.split(".")[-1],
                "FULLPATH": fullpath,
                "BASENAME": os.path.split(localpath)[-1],
            }
            self.files = pd.concat(
                [self.files, pd.DataFrame([filerow])], ignore_index=True
            )
        try:
            value = pd.read_csv(
                fullpath,
                header=None,
                sep="DONOTSEPARATEANYTHING *%magic%*",
                engine="python",
                skip_blank_lines=skip_blank_lines,
                skipinitialspace=skipinitialspace,
                comment=comment,
            ).iloc[0, 0]
        except pd.errors.EmptyDataError:
            value = ""
        if convert_numeric:
            value = parse_number(value)
            if not isinstance(value, str):
                self.data[localpath] = value
            else:
                # In case we are re-reading, we must
                # ensure there is no value present now:
                if localpath in self.data:
                    del self.data[localpath]
        else:
            self.data[localpath] = value
        return value

    def load_txt(self, localpath, convert_numeric=True, force_reread=False):
        """Parse a txt file with
        <key> <value>
        in each line.

        The txt file will be internalized in a dict and will be
        stored if the object is archived. Recommended file
        extension is 'txt'.

        Common usage is internalization of parameters.txt which
        happens by default, but this can be used for all txt files.

        The parsed data is returned as a dict. At the ensemble level
        the same function returns a dataframe.

        There is no get'er for the constructed data, access the
        class variable keyvaluedata directly, or rerun this function.
        (except for parameters.txt, for which there is a property
        called 'parameters')

        Values with spaces are not supported, this is similar
        to ERT's CSV_EXPORT1. Remainder string will be ignored silently.

        Args:
            localpath: path local the realization to the txt file
            convert_numeric: defaults to True, will try to parse
                all values as integers, if not, then floats, and
                strings as the last resort.
            force_reread: Force reread from file system. If
                False, repeated calls to this function will
                returned cached results.

        Returns:
            dict: Dictionary with the parsed values. Values will be returned as
                integers, floats or strings. If convert_numeric
                is False, all values are strings.
        """
        fullpath = os.path.abspath(os.path.join(self._origpath, localpath))
        if not os.path.exists(fullpath):
            raise IOError("File not found: " + fullpath)
        if fullpath in self.files["FULLPATH"].values and not force_reread:
            # Return cached version
            return self.data[localpath]
        if fullpath not in self.files["FULLPATH"].values:
            filerow = {
                "LOCALPATH": localpath,
                "FILETYPE": localpath.split(".")[-1],
                "FULLPATH": fullpath,
                "BASENAME": os.path.split(localpath)[-1],
            }
            self.files = pd.concat(
                [self.files, pd.DataFrame([filerow])], ignore_index=True
            )
        try:
            keyvalues = pd.read_csv(
                fullpath,
                sep=r"\s+",
                index_col=0,
                dtype=str,
                usecols=[0, 1],
                header=None,
            )[1].to_dict()
        except pd.errors.EmptyDataError:
            keyvalues = {}
        if convert_numeric:
            for key in keyvalues:
                keyvalues[key] = parse_number(keyvalues[key])
        self.data[localpath] = keyvalues
        return keyvalues

    def load_csv(self, localpath, convert_numeric=True, force_reread=False):
        """Parse a CSV file as a DataFrame

        Data will be stored as a DataFrame for later
        access or storage.

        Filename is relative to realization root.

        Args:
            localpath: path local the realization to the txt file
            convert_numeric: defaults to True, will try to parse
                all values as integers, if not, then floats, and
                strings as the last resort.
            force_reread: Force reread from file system. If
                False, repeated calls to this function will
                returned cached results.

        Returns:
            dataframe: The CSV file loaded. Empty dataframe
                if file is not present.
        """
        fullpath = os.path.abspath(os.path.join(self._origpath, localpath))
        if not os.path.exists(fullpath):
            raise IOError("File not found: " + fullpath)
        # Look for cached version
        if localpath in self.data and not force_reread:
            return self.data[localpath]
        # Check the file store, append if not there
        if localpath not in self.files["LOCALPATH"].values:
            filerow = {
                "LOCALPATH": localpath,
                "FILETYPE": localpath.split(".")[-1],
                "FULLPATH": fullpath,
                "BASENAME": os.path.split(localpath)[-1],
            }
            self.files = pd.concat(
                [self.files, pd.DataFrame([filerow])], ignore_index=True
            )
        try:
            if convert_numeric:
                # Trust that Pandas will determine sensible datatypes
                # faster than the convert_numeric() function
                dtype = None
            else:
                dtype = str
            dframe = pd.read_csv(fullpath, dtype=dtype)
            if "REAL" in dframe:
                dframe.rename(columns={"REAL": "REAL_ORIG"}, inplace=True)
                logger.warning(
                    (
                        "Loaded file %s already had the column REAL, "
                        "this was renamed to REAL_ORIG"
                    ),
                    fullpath,
                )
        except pd.errors.EmptyDataError:
            dframe = None  # or empty dataframe?

        # Store parsed data:
        self.data[localpath] = dframe
        return dframe

    def load_status(self):
        """Collects the contents of the STATUS files and return
        as a dataframe, with information from jobs.json added if
        available.

        Each row in the dataframe is a finished FORWARD_MODEL
        The STATUS files are parsed and information is extracted.
        Job duration is calculated, but jobs above 24 hours
        get incorrect durations.

        Returns:
            A dataframe with information from the STATUS files.
            Each row represents one job in one of the realizations.
        """

        statusfile = os.path.join(self._origpath, "STATUS")
        if not os.path.exists(statusfile):
            # This should not happen as long as __init__ requires STATUS
            # to be present.
            return pd.DataFrame()  # will be empty
        errorcolumns = ["error" + str(x) for x in range(0, 10)]

        status = pd.read_csv(
            statusfile,
            sep=r"\s+",
            skiprows=1,
            header=None,
            names=["FORWARD_MODEL", "colon", "STARTTIME", "dots", "ENDTIME"]
            + errorcolumns,
            dtype=str,
            engine="python",
            on_bad_lines="skip",
        )

        status.fillna("", inplace=True)

        errorjobs = status[errorcolumns[0]] != ""

        # Merge any error strings:
        error_string = (
            status.loc[errorjobs, errorcolumns]
            .astype(str)
            .apply(" ".join, axis=1)
            .apply(str.strip)
        )
        status["errorstring"] = np.nan
        status.loc[errorjobs, "errorstring"] = error_string
        status.drop(errorcolumns, axis=1, inplace=True)

        # Delete potential unwanted row
        status = status[~((status.FORWARD_MODEL == "LSF") & (status.colon == "JOBID:"))]

        if status.empty:
            logger.warning("No parseable data in STATUS")
            self.data["STATUS"] = status
            return status

        status = status.reset_index().drop("colon", axis=1).drop("dots", axis=1)

        # Index the jobs, this makes it possible to match with jobs.json:
        status.insert(0, "JOBINDEX", status.index.astype(int))
        status = status.drop("index", axis=1)
        # Calculate duration. Only Python 3.6 has time.fromisoformat().
        # Warning: Unpandaic code..
        durations = []
        for _, jobrow in status.iterrows():
            if not jobrow["ENDTIME"]:  # A job that is not finished.
                durations.append(np.nan)
            else:
                try:
                    hms = list(map(int, jobrow["STARTTIME"].split(":")))
                    start = datetime.combine(
                        date.today(), time(hour=hms[0], minute=hms[1], second=hms[2])
                    )
                    hms = list(map(int, jobrow["ENDTIME"].split(":")))
                    end = datetime.combine(
                        date.today(), time(hour=hms[0], minute=hms[1], second=hms[2])
                    )
                    # This works also when we have crossed 00:00:00.
                    # Jobs > 24 h will be wrong.
                    durations.append((end - start).seconds)
                except ValueError:
                    # We get where if STARTIME.split(':') does not contain
                    # integers only:
                    durations.append(np.nan)
        status["DURATION"] = durations

        # Augment data from jobs.json if that file is available:
        jsonfilename = os.path.join(self._origpath, "jobs.json")
        if jsonfilename and os.path.exists(jsonfilename):
            try:
                with open(jsonfilename) as file_handle:
                    jobsinfo = json.load(file_handle)
                jobsinfodf = pd.DataFrame(jobsinfo["jobList"])
                jobsinfodf["JOBINDEX"] = jobsinfodf.index.astype(int)
                # Outer merge means that we will also have jobs from
                # jobs.json that has not started (failed or perhaps
                # the jobs are still running on the cluster)
                status = status.merge(jobsinfodf, how="outer", on="JOBINDEX")
            except ValueError:
                logger.warning("Parsing file %s failed, skipping", jsonfilename)
        status.sort_values(["JOBINDEX"], ascending=True, inplace=True)
        self.data["STATUS"] = status
        return status

    def apply(self, callback, **kwargs):
        """Callback functionality

        A function handle can be supplied which will be executed on
        this realization. The function supplied *must* return
        a Pandas DataFrame. The function can accept an additional
        kwargs dictionary with extra information. Special keys
        in the kwargs data are 'realization', which will hold
        the current realization object. The key 'localpath' is
        also reserved for the use inside this apply(), as it
        is used for the name of the internalized data.

        If the key 'dumptofile' is a boolean and set to True,
        the resulting dataframe is also attempted written
        to disk using the supplied 'localpath'.

        Args:
            **kwargs (dict): which is supplied to the callbacked function,
                in which the key 'localpath' also points the the name
                used for data internalization.
        """

        if not kwargs:
            kwargs = {}
        if "realization" in kwargs:
            raise ValueError("Never supply realization= to apply()")
        kwargs["realization"] = self

        # Allow for calling functions which cannot take any
        # arguments:
        try:
            result = callback(kwargs)  # lgtm [py/call/wrong-arguments]
        except TypeError:
            result = callback()

        if not isinstance(result, pd.DataFrame):
            raise ValueError(
                "Returned value from applied " + "function must be a dataframe"
            )

        # Only internalize if 'localpath' is given
        if "localpath" in kwargs:
            self.data[kwargs["localpath"]] = result

        if "dumptodisk" in kwargs and kwargs["dumptodisk"]:
            if not kwargs["localpath"]:
                raise ValueError(
                    "localpath must be supplied when" + "dumptodisk is used"
                )
            fullpath = os.path.join(self.runpath(), kwargs["localpath"])
            if not os.path.exists(os.path.dirname(fullpath)):
                os.makedirs(os.path.dirname(fullpath))
            if os.path.exists(fullpath):
                os.unlink(fullpath)
            logger.info("Writing result of function call to %s", fullpath)
            result.to_csv(fullpath, index=False)
        return result

    def __getitem__(self, localpath):
        """Direct access to the realization data structure

        Calls get_df(localpath).
        """
        return self.get_df(localpath)

    def __delitem__(self, localpath):
        """Deletes components in the internal datastore.

        Silently ignores data that is not found.

        Args:
            localpath: string, fully qualified name of key
                (no shorthand as for get_df())
        """
        if localpath in self.keys():
            del self.data[localpath]

    def keys(self):
        """Access the keys of the internal data structure"""
        return self.data.keys()

    def get_df(self, localpath, merge=None):
        """Access the internal datastore which contains dataframes or dicts
        or scalars.

        The localpath argument can be shortened, as it will be
        looked up using the function shortcut2path()

        Args:
            localpath (str): the idenfier of the data requested
            merge (list or str): identifier/localpath of some data to be merged in,
                typically 'parameters.txt'. Will only work when return type is a
                dataframe. If list is supplied, order can matter.

        Returns:
            dataframe or dictionary.

        Raises:
            KeyError if data is not found.
            TypeError if data in localpath or merge is not of a mergeable type
        """
        fullpath = shortcut2path(self.keys(), localpath)
        if fullpath not in self.data.keys():
            raise KeyError("Could not find {}".format(localpath))
        data = self.data[shortcut2path(self.keys(), localpath)]
        if not isinstance(merge, list):
            merge = [merge]  # can still be None
        if merge and merge[0] is not None:
            # Strange things can happen when we do merges since
            # this function happily returns references to the internal
            # dataframes in the realization object. So ensure
            # we copy dataframes if any merging is about to happen.
            if isinstance(data, pd.DataFrame):
                data = data.copy()
            elif isinstance(data, dict):
                data = data.copy()
            elif isinstance(data, (str, int, float, np.number)):
                # Convert scalar data into something mergeable
                value = data
                data = {localpath: value}
            else:
                raise TypeError(
                    "Don't know how to merge data "
                    + "from {} of type {}".format(localpath, type(data))
                )
        for mergekey in merge:
            if mergekey is None:
                continue
            mergedata = self.get_df(mergekey)
            if isinstance(mergedata, dict):
                for key in mergedata:
                    # Add a column to the data for each dictionary
                    # key:
                    data[key] = mergedata[key]
            elif isinstance(mergedata, (str, int, float, np.number)):
                # Scalar data, use the mergekey as column
                data[mergekey] = mergedata
            elif isinstance(mergedata, pd.DataFrame):
                data = pd.merge(data, mergedata)
                # pd.MergeError will be raised here when this fails,
                # there must be common columns for this operation.
            else:
                raise TypeError(
                    "Don't know how to merge data "
                    + "from {} of type {}".format(mergekey, type(data))
                )
        return data

    def find_files(self, paths, metadata=None, metayaml=False):
        """Discover realization files. The files dataframe
        will be updated.

        Certain functionality requires up-front file discovery,
        e.g. ensemble archiving and ensemble arithmetic.

        CSV files for single use do not have to be discovered.

        Files containing double-dashes '--' indicate that the double
        dashes separate different component with meaning in the
        filename.  The components are extracted and put into
        additional columns "COMP1", "COMP2", etc..
        Filetype extension (after the last dot) will be removed
        from the last component.

        Args:
            paths: str or list of str with filenames (will be globbed)
                that are relative to the realization directory.
            metadata: dict with metadata to assign for the discovered
                files. The keys will be columns, and its values will be
                assigned as column values for the discovered files.
                During rediscovery of files, old metadata will be removed.
            metayaml: Additional possibility of adding metadata from
                associated yaml files. Yaml files to be associated to
                a specific discovered file can have an optional dot in
                front, and must end in .yml, added to the discovered filename.
                The yaml file will be loaded as a dict, and have its keys
                flattened using the separator '--'. Flattened keys are
                then used as column headers in the returned dataframe.

        Returns:
            A slice of the internalized dataframe corresponding
            to the discovered files (will be included even if it has
            been discovered earlier)
        """
        if isinstance(paths, str):
            paths = [paths]
        returnedslice = pd.DataFrame(
            columns=["FULLPATH", "FILETYPE", "LOCALPATH", "BASENAME"]
        )
        for searchpath in paths:
            globs = [
                f
                for f in glob.glob(os.path.join(self._origpath, searchpath))
                if os.path.isfile(f)
            ]
            for match in globs:
                absmatch = os.path.abspath(match)
                dirname = os.path.dirname(absmatch)
                basename = os.path.basename(match)
                filetype = match.split(".")[-1]
                filerow = {
                    "LOCALPATH": os.path.relpath(match, self._origpath),
                    "FILETYPE": filetype,
                    "FULLPATH": absmatch,
                    "BASENAME": basename,
                }
                # Look for and split basename based on double-dash '--'
                basename_noext = basename.replace("." + filetype, "")
                if "--" in basename_noext:
                    for compidx, comp in enumerate(basename_noext.split("--")):
                        filerow["COMP" + str(compidx + 1)] = comp

                if metayaml:
                    metadict = {}
                    yaml_candidates = [
                        "." + basename + ".yml",
                        basename + ".yml",
                        "." + basename + ".yaml",
                        basename + ".yaml",
                    ]
                    # We will only parse the first one found! You
                    # might be out of luck if you have multiple..
                    for cand in yaml_candidates:
                        if os.path.exists(os.path.join(dirname, cand)):
                            with open(os.path.join(dirname, cand)) as file_handle:
                                metadict = yaml.full_load(file_handle)
                        continue

                    # Flatten metadict:
                    metadict = flatten(metadict, sep="--")
                    for key, value in metadict.items():
                        if key not in filerow:
                            filerow[key] = value
                        else:
                            logger.warning(
                                "Cannot add key %s from yaml, key is in use. Skipping.",
                                key,
                            )

                # Delete this row if it already exists, determined by FULLPATH
                if absmatch in self.files["FULLPATH"].values:
                    self.files = self.files[self.files["FULLPATH"] != absmatch]
                if metadata:
                    filerow.update(metadata)
                self.files = pd.concat(
                    [self.files, pd.DataFrame([filerow])], ignore_index=True
                )
                returnedslice = pd.concat(
                    [returnedslice, pd.DataFrame([filerow])], ignore_index=True
                )
        return returnedslice

    @property
    def parameters(self):
        """Access the data obtained from parameters.txt

        Returns:
            dict with data from parameters.txt
        """
        return self.data["parameters.txt"]

    def get_eclfiles(self):
        """
        Return an ecl2df.EclFiles object to connect to the ecl2df package

        If autodiscovery, it will search for a DATA file in
        the standard location eclipse/model/...DATA.

        If you have multiple DATA files, you must discover
        the one you need explicitly before calling this function, example:

            >>> real = ScratchRealization("myrealpath")
            >>> real.find_files("eclipse/model/MYMODELPREDICTION.DATA")

        Returns:
            ecl2df.EclFiles. None if nothing found
        """
        if not HAVE_ECL2DF:
            logger.warning("ecl2df not installed. Skipping")
            return None
        data_file_row = self.files[self.files["FILETYPE"] == "DATA"]
        data_filename = None
        if len(data_file_row) == 1:
            data_filename = data_file_row["FULLPATH"].values[0]
        elif self._autodiscovery:
            data_fileguess = os.path.join(self._origpath, "eclipse/model", "*.DATA")
            data_filenamelist = glob.glob(data_fileguess)
            if not data_filenamelist:
                return None  # No filename matches *DATA
            if len(data_filenamelist) > 1:
                logger.warning(
                    (
                        "Multiple DATA files found, "
                        "consider turning off auto-discovery"
                    )
                )
            data_filename = data_filenamelist[0]
            self.find_files(data_filename)
        else:
            # There is no DATA file to be found.
            logger.warning("No DATA file found!")
            return None
        if not os.path.exists(data_filename):
            return None
        return ecl2df.EclFiles(data_filename)

    def get_eclsum(self, cache=True, include_restart=True):
        """
        Fetch the Eclipse Summary file from the realization
        and return as a libecl EclSum object

        Unless the UNSMRY file has been discovered, it will
        pick the file from the glob `eclipse/model/*UNSMRY`,
        as long as autodiscovery is not turned off when
        the realization object was initialized.

        If you have multiple UNSMRY files in eclipse/model
        turning off autodiscovery is strongly recommended.

        Arguments:
            cache: boolean indicating whether we should keep an
                object reference to the EclSum object. Set to
                false if you need to conserve memory.
            include_restart: boolean sent to libecl for whether restart
                files should be traversed.

        Returns:
            EclSum: object representing the summary file. None if
                nothing was found.
        """
        if cache and self._eclsum:  # Return cached object if available
            if self._eclsum_include_restart == include_restart:
                return self._eclsum

        unsmry_file_row = self.files[self.files.FILETYPE == "UNSMRY"]
        unsmry_filename = None
        if len(unsmry_file_row) == 1:
            unsmry_filename = unsmry_file_row.FULLPATH.values[0]
        elif self._autodiscovery:
            unsmry_fileguess = os.path.join(self._origpath, "eclipse/model", "*.UNSMRY")
            unsmry_filenamelist = glob.glob(unsmry_fileguess)
            if not unsmry_filenamelist:
                return None  # No filename matches
            if len(unsmry_filenamelist) > 1:
                logger.warning(
                    "Multiple UNSMRY files found, consider turning off auto-discovery"
                )
            unsmry_filename = unsmry_filenamelist[0]
            self.find_files(unsmry_filename)
        else:
            # There is no UNSMRY file to be found.
            return None

        if not os.path.exists(unsmry_filename):
            return None
        try:
            eclsum = Summary(
                unsmry_filename, lazy_load=False, include_restart=include_restart
            )
        except IOError:
            # This can happen if there is something wrong with the file
            # or if SMSPEC is missing.
            logger.warning("Failed to create summary instance from %s", unsmry_filename)
            return None

        if cache:
            self._eclsum = eclsum
            self._eclsum_include_restart = include_restart

        return eclsum

    def load_smry(
        self,
        time_index="raw",
        column_keys=None,
        cache_eclsum=None,
        start_date=None,
        end_date=None,
        include_restart=True,
    ):
        """Produce dataframe from Summary data from the realization

        When this function is called, the dataframe will be
        internalized.  Internalization of summary data in a
        realization object supports different time_index, but there is
        no handling of multiple sets of column_keys. The cached data
        will be called

          'share/results/tables/unsmry--<time_index>.csv'

        where <time_index> is among 'yearly', 'monthly', 'daily', 'first',
        'last' or 'raw' (meaning the raw dates in the SMRY file), depending
        on the chosen time_index. If a custom time_index (list
        of datetime) was supplied, <time_index> will be called 'custom'.

        Wraps resdata.summary.Summary.pandas_frame()

        See also get_smry()

        Args:
            time_index: string indicating a resampling frequency,
               'yearly', 'monthly', 'daily', 'first', 'last' or 'raw', the
               latter will return the simulated report steps (also default).
               If a list of DateTime is supplied, data will be resampled
               to these.
            column_keys: list of column key wildcards. None means everything.
            cache_eclsum: boolean for whether to keep the loaded EclSum
                object in memory after data has been loaded.
            start_date: str or date with first date to include.
                Dates prior to this date will be dropped, supplied
                start_date will always be included. Overridden if time_index
                is 'first' or 'last'.
            end_date: str or date with last date to be included.
                Dates past this date will be dropped, supplied
                end_date will always be included. Overridden if time_index
                is 'first' or 'last'.
            include_restart: boolean sent to libecl for whether restart
                files should be traversed.

        Returns:
            DataFrame with summary keys as columns and dates as indices.
                Empty dataframe if no summary is available or column
                keys do not exist.
            DataFrame: with summary keys as columns and dates as indices.
                Empty dataframe if no summary is available.
        """
        if cache_eclsum is not None:
            warnings.warn(
                (
                    "cache_eclsum option to load_smry() is deprecated and "
                    "will be removed in fmu-ensemble v2.0.0"
                ),
                FutureWarning,
            )
        else:
            cache_eclsum = True

        if not self.get_eclsum(cache=cache_eclsum):
            # Return empty, but do not store the empty dataframe in self.data
            return pd.DataFrame()
        time_index_path = time_index
        if time_index == "raw":
            time_index_arg = None
        elif isinstance(time_index, str):
            # Note: This call will recache the smry object.
            time_index_arg = self.get_smry_dates(
                freq=time_index,
                start_date=start_date,
                end_date=end_date,
                include_restart=include_restart,
            )
        elif isinstance(time_index, (list, np.ndarray)):
            time_index_arg = time_index
            time_index_path = "custom"
        elif time_index is None:
            time_index_path = "raw"
            time_index_arg = time_index
        else:
            raise TypeError("'time_index' has to be a string, a list or None")

        if not isinstance(column_keys, list):
            column_keys = [column_keys]

        # Do the actual work:
        dframe = self.get_eclsum(
            cache=cache_eclsum, include_restart=include_restart
        ).pandas_frame(time_index_arg, column_keys)
        dframe = dframe.reset_index()
        dframe.rename(columns={"index": "DATE"}, inplace=True)

        # Cache the result:
        localpath = "share/results/tables/unsmry--" + time_index_path + ".csv"
        self.data[localpath] = dframe

        # Do this to ensure that we cut the rope to the EclSum object
        # Can be critical for garbage collection
        if not cache_eclsum:
            self._eclsum = None
        return dframe

    def get_smry(
        self,
        time_index=None,
        column_keys=None,
        cache_eclsum=None,
        start_date=None,
        end_date=None,
        include_restart=True,
    ):
        """Wrapper for EclSum.pandas_frame

        This gives access to the underlying data on disk without
        touching internalized dataframes.

        Arguments:
            time_index: string indicating a resampling frequency,
               'yearly', 'monthly', 'daily', 'first', 'last' or 'raw', the
               latter will return the simulated report steps (also default).
               If a list of DateTime is supplied, data will be resampled
               to these. If a date in ISO-8601 format is supplied, that is
               used as a single date.
            column_keys: list of column key wildcards. None means everything.
            cache_eclsum: boolean for whether to keep the loaded EclSum
                object in memory after data has been loaded.
            start_date: str or date with first date to include.
                Dates prior to this date will be dropped, supplied
                start_date will always be included. Overridden if time_index
                is 'first' or 'last'.
            end_date: str or date with last date to be included.
                Dates past this date will be dropped, supplied
                end_date will always be included. Overridden if time_index
                is 'first' or 'last'.

        Returns empty dataframe if there is no summary file, or if the
        column_keys are not existing.
        """

        if cache_eclsum is not None:
            warnings.warn(
                (
                    "cache_eclsum option to get_smry() is deprecated and "
                    "will be removed in fmu-ensemble v2.0.0"
                ),
                FutureWarning,
            )
        else:
            cache_eclsum = True

        if not isinstance(column_keys, list):
            column_keys = [column_keys]
        if isinstance(time_index, str) and time_index == "raw":
            time_index_arg = None
        elif isinstance(time_index, str):
            try:
                parseddate = dateutil.parser.isoparse(time_index)
                time_index_arg = [parseddate]
            except ValueError:
                time_index_arg = self.get_smry_dates(
                    freq=time_index,
                    start_date=start_date,
                    end_date=end_date,
                    include_restart=include_restart,
                )
        elif time_index is None or isinstance(time_index, (list, np.ndarray)):
            time_index_arg = time_index
        else:
            raise TypeError("'time_index' has to be a string, a list or None")
        if self.get_eclsum(cache=cache_eclsum, include_restart=include_restart):
            try:
                dataframe = self.get_eclsum(
                    cache=cache_eclsum, include_restart=include_restart
                ).pandas_frame(time_index_arg, column_keys)
            except ValueError:
                # We get here if we have requested non-existing column keys
                return pd.DataFrame()
            if not cache_eclsum:
                # Ensure EclSum object can be garbage collected
                self._eclsum = None
            return dataframe
        return pd.DataFrame()

    def get_smry_meta(self, column_keys=None):
        """
        Provide metadata for summary data vectors.

        A dictionary indexed by summary vector names is returned, and each
        value is another dictionary with potentially the metadata types:
        * unit (string)
        * is_total (bool)
        * is_rate (bool)
        * is_historical (bool)
        * get_num (int) (only provided if not None)
        * keyword (str)
        * wgname (str or None)

        Args:
            column_keys: List or str of column key wildcards
        """
        column_keys = self._glob_smry_keys(column_keys)
        meta = {}
        eclsum = self.get_eclsum()
        for col in column_keys:
            meta[col] = {}
            meta[col]["unit"] = eclsum.unit(col)
            meta[col]["is_total"] = eclsum.is_total(col)
            meta[col]["is_rate"] = eclsum.is_rate(col)
            meta[col]["is_historical"] = eclsum.smspec_node(col).is_historical()
            meta[col]["keyword"] = eclsum.smspec_node(col).keyword
            meta[col]["wgname"] = eclsum.smspec_node(col).wgname
            num = eclsum.smspec_node(col).get_num()
            if num is not None:
                meta[col]["get_num"] = num
        return meta

    def _glob_smry_keys(self, column_keys):
        """Utility function for globbing column names

        Use this to expand 'F*' to the list of Eclipse summary
        vectors matching.

        Args:
            column_keys: str or list of strings with patterns

        Returns:
            list of strings. Empty list if no summary loaded.
        """
        if self.get_eclsum() is None:
            logger.warning(
                (
                    "Calling _glob_smry_keys without loaded or found summary file "
                    "returns empty list"
                )
            )
            return []
        if not isinstance(column_keys, list):
            column_keys = [column_keys]
        keys = set()
        for key in column_keys:
            if isinstance(key, str):
                keys = keys.union(set(self._eclsum.keys(key)))
        return list(keys)

    def get_volumetric_rates(self, column_keys=None, time_index=None, time_unit=None):
        """Compute volumetric rates from cumulative summary vectors

        See :meth:`fmu.ensemble.util.compute_volumetric_rates`
        """
        return compute_volumetric_rates(self, column_keys, time_index, time_unit)

    def get_smryvalues(self, props_wildcard=None):
        """
        Fetch selected vectors from Eclipse Summary data.

        Args:
            props_wildcard : string or list of strings with vector
                wildcards
        Returns:
            a dataframe with values. Raw times from UNSMRY.
            Empty dataframe if no summary file data available
        """
        warnings.warn(
            (
                "realization.get_smryvalues() is deprecated and "
                "will be removed in fmu-ensemble v2.0.0. Process "
                "columns from get_smry() instead."
            ),
            FutureWarning,
        )

        if not self._eclsum:  # check if it is cached
            self.get_eclsum()

        if not self._eclsum:
            return pd.DataFrame()

        props = self._glob_smry_keys(props_wildcard)

        if "numpy_vector" in dir(self._eclsum):
            data = {
                prop: self._eclsum.numpy_vector(prop, report_only=False)
                for prop in props
            }
        else:  # get_values() is deprecated in newer libecl
            data = {
                prop: self._eclsum.get_values(prop, report_only=False) for prop in props
            }
        dates = self._eclsum.get_dates(report_only=False)
        return pd.DataFrame(data=data, index=dates)

    def get_smry_dates(
        self,
        freq="monthly",
        normalize=True,
        start_date=None,
        end_date=None,
        include_restart=True,
    ):
        """Return list of datetimes available in the realization

        Args:
            freq: string denoting requested frequency for
                the returned list of datetime. 'report' will
                yield the sorted union of all valid timesteps for
                all realizations. Other valid options are
                'daily', 'weekly', 'monthly' and 'yearly'.
                'first' will give out the first date (minimum) and
                'last' will give out the last date (maximum),
                both as lists with one element.
            normalize: Whether to normalize backwards at the start
                and forwards at the end to ensure the raw
                date range is covered.
            start_date: str or date with first date to include
                Dates prior to this date will be dropped, supplied
                start_date will always be included. Overrides
                normalized dates. Overridden if freq is 'first'
                or 'last'.
            end_date: str or date with last date to be included.
                Dates past this date will be dropped, supplied
                end_date will always be included. Overrides
                normalized dates. Overridden if freq is 'first'
                or 'last'.
        Returns:
            list of datetimes. None if no summary data is available.
        """
        eclsum = self.get_eclsum(include_restart=include_restart)
        if not eclsum:
            return None
        return unionize_smry_dates(
            [eclsum.dates], freq, normalize, start_date, end_date
        )

    def contains(self, localpath, **kwargs):
        """Boolean function for asking the realization for presence
        of certain data types and possibly data values.

        Args:
            localpath: string pointing to the data for which the query
                applies. If no other arguments, only realizations containing
                this data key is kept.
            key: A certain key within a realization dictionary that is
                required to be present. If a value is also provided, this
                key must be equal to this value. If localpath is not
                a dictionary, this will raise a ValueError
            value: The value a certain key must equal. Floating point
                comparisons are not robust. Only relevant for dictionaries
            column: Name of a column in tabular data. If columncontains is
                not specified, this means that this column must be present
            columncontains:
                A value that the specific column must include.

        Returns:
            boolean: True if the data is present and fulfilling any
            criteria.
        """
        kwargs.pop("inplace", 0)
        localpath = shortcut2path(self.keys(), localpath)
        if localpath not in self.keys():
            return False
        if not kwargs:
            return localpath in self.keys()
        if isinstance(self.data[localpath], dict):
            if "key" in kwargs and "value" not in kwargs:
                return kwargs["key"] in self.data[localpath]
        if isinstance(self.data[localpath], pd.DataFrame):
            if "key" in kwargs:
                raise ValueError("Don't use key for tabular data")
            if "value" in kwargs:
                raise ValueError("Don't use value for tabular data")
            if "column" in kwargs and "columncontains" not in kwargs:
                # Only asking for column presence
                return kwargs["column"] in self.data[localpath].columns
            if "column" in kwargs and "columncontains" in kwargs:
                # If we are dealing with the DATE column,
                # convert everything to pandas datatime64 for comparisons,
                # otherwise we revert to simpler check.
                if kwargs["column"] == "DATE":
                    return (
                        pd.to_datetime(dateutil.parser.parse(kwargs["columncontains"]))
                        == pd.to_datetime(self.data[localpath][kwargs["column"]])
                    ).any()
                return (
                    kwargs["columncontains"]
                    in self.data[localpath][kwargs["column"]].values
                )

        if "key" in kwargs and "value" in kwargs:
            if isinstance(kwargs["value"], str):
                if kwargs["key"] in self.data[localpath]:
                    return str(self.data[localpath][kwargs["key"]]) == kwargs["value"]
                return False
            # non-string, then don't convert the internalized data
            return self.data[localpath][kwargs["key"]] == kwargs["value"]
        raise ValueError("Wrong arguments to contains()")

    def drop(self, localpath, **kwargs):
        """Delete elements from internalized data.

        Shortcuts are allowed for localpath. If the data pointed to is
        a DataFrame, you can delete columns, or rows containing certain
        elements

        If the data pointed to is a dictionary, keys can be deleted.

        Args:
            localpath: string, path to internalized data. If no other options
                are supplied, that dataset is deleted in its entirety
            column: string with a column name to drop. Only for dataframes
            columns: list of strings with column names to delete
            rowcontains: rows where one column contains this string will be
                dropped. The comparison is on strings only, and all cells in
                the dataframe is converted to strings for the comparison.
                Thus it might work on dates, but be careful with numbers.
            key: string with a keyname in a dictionary. Will not work for
                dataframes
            keys: list of strings of keys to delete from a dictionary
        """
        fullpath = shortcut2path(self.keys(), localpath)
        if fullpath not in self.keys():
            raise ValueError("%s not found" % localpath)

        data = self.data[fullpath]

        if not kwargs:
            # This will remove the entire dataset
            self.data.pop(fullpath, None)

        if isinstance(data, pd.DataFrame):
            if "column" in kwargs:
                data.drop(labels=kwargs["column"], axis="columns", inplace=True)
            if "columns" in kwargs:
                data.drop(labels=kwargs["columns"], axis="columns", inplace=True)
            if "rowcontains" in kwargs:
                # Construct boolean series for those rows that have a match
                boolseries = (data.astype(str) == str(kwargs["rowcontains"])).any(
                    axis="columns"
                )
                self.data[fullpath] = data[~boolseries]
        if isinstance(data, dict):
            if "keys" in kwargs:
                for key in kwargs["keys"]:
                    data.pop(key, None)
            if "key" in kwargs:
                data.pop(kwargs["key"], None)

    def __repr__(self):
        """Represent the realization. Show only the last part of the path"""
        pathsummary = self._origpath[-50:]
        if self.index is not None:
            indexstr = str(self.index)
        else:
            indexstr = "Error"
        return "<Realization, index={}, path=...{}>".format(indexstr, pathsummary)

    def __sub__(self, other):
        """Substract another realization from this"""
        result = RealizationCombination(ref=self, sub=other)
        return result

    def __add__(self, other):
        """Add another realization from this"""
        result = RealizationCombination(ref=self, add=other)
        return result

    def __mul__(self, other):
        """Scale this realization by a scalar value"""
        result = RealizationCombination(ref=self, scale=float(other))
        return result

    def __rsub__(self, other):
        """Add another realization from this"""
        result = RealizationCombination(ref=self, sub=other)
        return result

    def __radd__(self, other):
        """Substract another realization from this"""
        result = RealizationCombination(ref=self, add=other)
        return result

    def __rmul__(self, other):
        """Scale this realization by a scalar value"""
        result = RealizationCombination(ref=self, scale=float(other))
        return result

    def get_init(self):
        """
        :returns: init file of the realization.
        """
        warnings.warn(
            (
                "realization.get_init() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        init_file_row = self.files[self.files.FILETYPE == "INIT"]
        init_filename = None
        if len(init_file_row) == 1:
            init_filename = init_file_row.FULLPATH.values[0]
        else:
            init_fileguess = os.path.join(self._origpath, "eclipse/model", "*.INIT")
            init_filenamelist = glob.glob(init_fileguess)
            if not init_filenamelist:
                return None  # No filename matches
            init_filename = init_filenamelist[0]
        if not os.path.exists(init_filename):
            return None

        if not self._eclinit:
            self._eclinit = ResdataFile(init_filename, flags=FileMode.CLOSE_STREAM)
        return self._eclinit

    def get_unrst(self):
        """
        :returns: restart file of the realization.
        """
        warnings.warn(
            (
                "realization.get_unrst() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        unrst_file_row = self.files[self.files.FILETYPE == "UNRST"]
        unrst_filename = None
        if len(unrst_file_row) == 1:
            unrst_filename = unrst_file_row.FULLPATH.values[0]
        else:
            unrst_fileguess = os.path.join(self._origpath, "eclipse/model", "*.UNRST")
            unrst_filenamelist = glob.glob(unrst_fileguess)
            if not unrst_filenamelist:
                return None  # No filename matches
            unrst_filename = unrst_filenamelist[0]
        if not os.path.exists(unrst_filename):
            return None
        if not self._eclunrst:
            self._eclunrst = ResdataFile(unrst_filename, flags=FileMode.CLOSE_STREAM)
        return self._eclunrst

    def get_grid_index(self, active_only):
        """
        Return the grid index in a pandas dataframe.
        """
        warnings.warn(
            (
                "realization.get_grid_index() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if self.get_grid():
            return self.get_grid().export_index(active_only=active_only)
        logger.warning("No GRID file in realization %s", self)

    def get_grid_corners(self, grid_index):
        """Return a dataframe with the the x, y, z for the
        8 grid corners of corner point cells"""
        warnings.warn(
            (
                "realization.get_grid_corners() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if self.get_grid():
            corners = self.get_grid().export_corners(grid_index)
            columns = [
                "x1",
                "y1",
                "z1",
                "x2",
                "y2",
                "z2",
                "x3",
                "y3",
                "z3",
                "x4",
                "y4",
                "z4",
                "x5",
                "y5",
                "z5",
                "x6",
                "y6",
                "z6",
                "x7",
                "y7",
                "z7",
                "x8",
                "y8",
                "z8",
            ]

            return pd.DataFrame(data=corners, columns=columns)
        else:
            logger.warning("No GRID file in realization %s", self)

    def get_grid_centre(self, grid_index):
        """Return the grid centre of corner-point-cells, x, y and z
        in distinct columns"""
        warnings.warn(
            (
                "realization.get_grid_centre() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if self.get_grid():
            grid_cell_centre = self.get_grid().export_position(grid_index)
            return pd.DataFrame(
                data=grid_cell_centre, columns=["cell_x", "cell_y", "cell_z"]
            )
        else:
            logger.warning("No GRID file in realization %s", self)

    def get_grid(self):
        """
        :returns: grid file of the realization.
        """
        warnings.warn(
            (
                "realization.get_grid() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        grid_file_row = self.files[self.files.FILETYPE == "EGRID"]
        grid_filename = None
        if len(grid_file_row) == 1:
            grid_filename = grid_file_row.FULLPATH.values[0]
        else:
            grid_fileguess = os.path.join(self._origpath, "eclipse/model", "*.EGRID")
            grid_filenamelist = glob.glob(grid_fileguess)
            if not grid_filenamelist:
                return None  # No filename matches
            grid_filename = grid_filenamelist[0]
        if not os.path.exists(grid_filename):
            return None
        if not self._eclgrid:
            self._eclgrid = Grid(grid_filename)
        return self._eclgrid

    @property
    def global_size(self):
        """
        :returns: Number of cells in the realization.
        """
        warnings.warn(
            (
                "realization.get_grid() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if self.get_grid() is not None:
            return self.get_grid().get_global_size()

    @property
    def actnum(self):
        """
        :returns: EclKw of ints showing which cells are active,
            Active cells are given value 1, while
            inactive cells have value 1.
        """
        warnings.warn(
            (
                "realization.get_grid() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if not self._actnum and self.get_init() is not None:
            self._actnum = self.get_init()["PORV"][0].create_actnum()
        return self._actnum

    @property
    def report_dates(self):
        """
        :returns: List of DateTime.DateTime for which values are reported.
        """
        warnings.warn(
            (
                "realization.get_grid() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if self.get_unrst() is not None:
            return self.get_unrst().report_dates

    def get_global_init_keyword(self, prop):
        """
        :param prop: A name of a keyword in the realization's init file.
        :returns: The EclKw of given name. Length is global_size.
            non-active cells are given value 0.
        """
        warnings.warn(
            (
                "realization.get_global_init_keyword() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if self.get_init() is not None:
            return self.get_init()[prop][0].scatter_copy(self.actnum)

    def get_global_unrst_keyword(self, prop, report):
        """
        :param prop: A name of a keyword in the realization's restart file.
        :returns: The EclKw of given name. Length is global_size.
            non-active cells are given value 0.
        """
        warnings.warn(
            (
                "realization.get_global_unrst_keyword() is deprecated and "
                "will be removed in later versions."
            ),
            FutureWarning,
        )
        if self.get_unrst() is not None:
            return self.get_unrst()[prop][report].scatter_copy(self.actnum)



"""Testing fmu-ensemble, virtualized ensembles"""

import logging
import os
from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from fmu.ensemble import ScratchEnsemble, VirtualEnsemble

try:
    # pylint: disable=unused-import
    import pyarrow  # noqa

    HAVE_PYARROW = True
except ImportError:
    HAVE_PYARROW = False

logger = logging.getLogger(__name__)


def test_virtualensemble():
    """Test the properties of a virtualized ScratchEnsemble"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    manifest = {
        "what": "A test ensemble for pytest usage",
        "coordinate_system": "The correct one",
    }

    reekensemble = ScratchEnsemble(
        "reektest",
        testdir + "/data/testensemble-reek001/" + "realization-*/iter-0",
        manifest=manifest,
    )
    reekensemble.load_smry(time_index="yearly", column_keys=["F*"])
    reekensemble.load_smry(column_keys=["FOPT", "FOIP"])
    reekensemble.load_smry(
        column_keys=["FGPT"],
        time_index=[
            datetime.strptime(strdate, "%Y-%m-%d %H:%M:%S")
            for strdate in ["2000-05-03 23:15:00", "2002-04-02 15:34:23"]
        ],
    )
    reekensemble.load_scalar("npv.txt")
    reekensemble.load_txt("outputs.txt")
    vens = reekensemble.to_virtual()

    assert "coordinate_system" in vens.manifest

    # Overwrite the manifest:
    vens.manifest = {"foo": "bar"}
    assert "foo" in vens.manifest
    assert "coordinate_system" not in vens.manifest

    # Check that we have data for 5 realizations
    assert len(vens["unsmry--yearly"]["REAL"].unique()) == 5
    assert len(vens["unsmry--raw"]["REAL"].unique()) == 5
    assert len(vens["unsmry--custom"]["REAL"].unique()) == 5
    assert len(vens["parameters.txt"]) == 5

    assert not vens.lazy_keys()

    # This is the dataframe of discovered files in the ScratchRealization
    assert isinstance(vens["__files"], pd.DataFrame)
    assert not vens["__files"].empty

    assert "REAL" in vens["STATUS"].columns

    # Check shorthand functionality:
    assert (
        vens.shortcut2path("unsmry--yearly")
        == "share/results/tables/unsmry--yearly.csv"
    )
    assert (
        vens.shortcut2path("unsmry--yearly.csv")
        == "share/results/tables/unsmry--yearly.csv"
    )

    assert "npv.txt" in vens.keys()
    assert len(vens["npv.txt"]) == 5  # includes the 'error!' string in real4
    assert "outputs.txt" in vens.keys()
    assert len(vens["outputs.txt"]) == 4

    # Check that get_smry() works
    # (here is with no interpolation necessary)
    fopt = vens.get_smry(column_keys=["FOPT"], time_index="yearly")
    assert "FOPT" in fopt.columns
    assert "DATE" in fopt.columns
    assert "REAL" in fopt.columns
    assert "FGPT" not in fopt.columns
    assert len(fopt) == 25

    # assert len(monthly_smry))==
    raw_smry = vens.get_smry(time_index="raw")
    pd.testing.assert_series_equal(
        vens.get_smry(time_index="first")["FOIP"].reset_index(drop=True),
        raw_smry[raw_smry["DATE"] == min(raw_smry["DATE"])]["FOIP"].reset_index(
            drop=True
        ),
    )
    pd.testing.assert_series_equal(
        vens.get_smry(time_index="last")["FOIP"].reset_index(drop=True),
        raw_smry[raw_smry["DATE"] == max(raw_smry["DATE"])]["FOIP"].reset_index(
            drop=True
        ),
    )

    # Check that we can default get_smry()
    alldefaults = vens.get_smry()
    # This should glob to all columns, and monthly time frequency
    # The 'monthly' is interpolated from the 'raw', as it is most likely
    # finer resolution than 'yearly'
    assert len(alldefaults) == 185
    assert len(alldefaults.columns) == 4
    # Check that monthly behaves the same way as default
    monthly_smry = vens.get_smry(time_index="monthly")
    assert len(monthly_smry) == 185
    assert len(monthly_smry.columns) == 4

    # Check that get_smry(time_index='raw')==get_smry(time_index=None)
    pd.testing.assert_series_equal(
        vens.get_smry(time_index="raw")["FOPT"].reset_index(drop=True),
        vens.get_smry(time_index=None)["FOPT"].reset_index(drop=True),
    )

    # Check that the custom smry has two dates
    assert len(vens.get_smry(time_index="custom")["DATE"].unique()) == 2

    # Eclipse summary vector statistics for a given ensemble
    df_stats = vens.get_smry_stats(column_keys=["FOPR", "FGPR"], time_index="yearly")
    assert isinstance(df_stats, pd.DataFrame)
    assert len(df_stats.columns) == 2
    assert isinstance(df_stats["FOPR"]["mean"], pd.Series)
    assert len(df_stats["FOPR"]["mean"]) == 5

    # Check webviz requirements for dataframe
    stats = df_stats.index.levels[0]
    assert "minimum" in stats
    assert "maximum" in stats
    assert "p10" in stats
    assert "p90" in stats
    assert "mean" in stats
    assert df_stats["FOPR"]["minimum"].iloc[-2] < df_stats["FOPR"]["maximum"].iloc[-2]

    # Test virtrealization retrieval:
    vreal = vens.get_realization(2)
    assert len(vreal.keys()) == len(vens.keys())
    assert set(vreal.keys()) == set(vens.keys())  # Order is not preserved

    # Test realization removal:
    vens.remove_realizations(3)
    assert len(vens.parameters["REAL"].unique()) == 4
    assert len(vens) == 4
    vens.remove_realizations(3)  # This will give warning
    assert len(vens.parameters["REAL"].unique()) == 4
    assert len(vens["unsmry--yearly"]["REAL"].unique()) == 4
    assert len(vens) == 4

    # Test data removal:
    vens.remove_data("parameters.txt")
    assert "parameters.txt" not in vens.keys()
    vens.remove_data("bogus")  # This should only give warning

    # Test data addition. It should(?) work also for earlier nonexisting
    vens.append(
        "betterdata",
        pd.DataFrame(
            {
                "REAL": [0, 1, 2, 3, 4, 5, 6, 80],
                "NPV": [1000, 2000, 1500, 2300, 6000, 3000, 800, 9],
            }
        ),
    )
    assert "betterdata" in vens.keys()
    assert "REAL" in vens["betterdata"].columns
    assert "NPV" in vens["betterdata"].columns

    assert vens.get_realization(3).get_df("betterdata")["NPV"] == 2300
    assert vens.get_realization(0).get_df("betterdata")["NPV"] == 1000
    assert vens.get_realization(1).get_df("betterdata")["NPV"] == 2000
    assert vens.get_realization(2).get_df("betterdata")["NPV"] == 1500
    assert vens.get_realization(80).get_df("betterdata")["NPV"] == 9

    with pytest.raises(ValueError):
        vens.get_realization(9999)

    assert vens.shortcut2path("betterdata") == "betterdata"
    assert vens.agg("min").get_df("betterdata")["NPV"] == 9
    assert vens.agg("max").get_df("betterdata")["NPV"] == 6000
    assert (
        vens.agg("min").get_df("betterdata")["NPV"]
        < vens.agg("p07").get_df("betterdata")["NPV"]
    )
    assert (
        vens.agg("p05").get_df("betterdata")["NPV"]
        < vens.agg("p55").get_df("betterdata")["NPV"]
    )
    assert (
        vens.agg("p46").get_df("betterdata")["NPV"]
        < vens.agg("max").get_df("betterdata")["NPV"]
    )

    assert "REAL" not in vens.agg("min")["STATUS"].columns

    # Betterdata should be returned as a dictionary
    # (it is returned from a virtualrealization object)
    assert isinstance(vens.agg("min").get_df("betterdata"), dict)


def test_todisk(tmpdir):
    """Test that we can write VirtualEnsembles to the filesystem in a
    retrievable manner"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")
    reekensemble = ScratchEnsemble(
        "reektest",
        testdir + "/data/testensemble-reek001/" + "realization-*/iter-0",
        manifest={"foo": "bar.com"},
    )
    reekensemble.load_smry(time_index="monthly", column_keys="*")
    reekensemble.load_smry(time_index="daily", column_keys="*")
    reekensemble.load_smry(time_index="yearly", column_keys="F*")
    reekensemble.load_scalar("npv.txt")
    reekensemble.load_txt("outputs.txt")
    vens = reekensemble.to_virtual()
    assert "foo" in vens.manifest

    tmpdir.chdir()

    vens.to_disk("vens_dumped", delete=True)
    assert len(vens) == len(reekensemble)

    fromdisk = VirtualEnsemble(fromdisk="vens_dumped")
    assert "foo" in fromdisk.manifest

    # Same number of realizations:
    assert len(fromdisk) == len(vens)

    # Should have all the same keys,
    # but change of order is fine
    assert set(vens.keys()) == set(fromdisk.keys())

    for frame in vens.keys():
        if frame == "STATUS":
            continue

        assert (vens.get_df(frame).columns == fromdisk.get_df(frame).columns).all()

        # Columns that only contains NaN will not have their
        # type preserved, this is too much to ask for, especially
        # with CSV files. So we drop columns with NaN
        virtframe = vens.get_df(frame).dropna(axis="columns")
        diskframe = fromdisk.get_df(frame).dropna(axis="columns")

        # It would be nice to be able to use pd.Dataframe.equals,
        # but it is too strict, as columns with mixed type number/strings
        # will easily be wrong.

        for column in set(virtframe.columns).intersection(set(diskframe.columns)):
            if object in (virtframe[column].dtype, diskframe[column].dtype):
                # Ensure we only compare strings when working with object dtype
                assert (
                    virtframe[column].astype(str).equals(diskframe[column].astype(str))
                )
            else:
                pd.testing.assert_series_equal(virtframe[column], diskframe[column])

    fromdisk.to_disk("vens_double_dumped", delete=True)
    # Here we could check filesystem equivalence if we want.

    vens.to_disk("vens_dumped_csv", delete=True, dumpparquet=False)
    fromcsvdisk = VirtualEnsemble(fromdisk="vens_dumped_csv")
    lazyfromdisk = VirtualEnsemble(fromdisk="vens_dumped_csv", lazy_load=True)
    assert set(vens.keys()) == set(fromcsvdisk.keys())
    assert set(vens.keys()) == set(lazyfromdisk.keys())
    assert "OK" in lazyfromdisk.lazy_frames.keys()
    assert "OK" not in lazyfromdisk.data.keys()
    assert len(fromcsvdisk.get_df("OK")) == len(lazyfromdisk.get_df("OK"))
    assert "OK" not in lazyfromdisk.lazy_frames.keys()
    assert "OK" in lazyfromdisk.data.keys()
    assert len(fromcsvdisk.parameters) == len(lazyfromdisk.parameters)
    assert len(fromcsvdisk.get_df("unsmry--yearly")) == len(
        lazyfromdisk.get_df("unsmry--yearly")
    )

    if HAVE_PYARROW:
        vens.to_disk("vens_dumped_parquet", delete=True, dumpcsv=False)
        fromparquetdisk = VirtualEnsemble()
        fromparquetdisk.from_disk("vens_dumped_parquet")
        assert set(vens.keys()) == set(fromparquetdisk.keys())

        fromparquetdisk2 = VirtualEnsemble()
        fromparquetdisk2.from_disk("vens_dumped_parquet", fmt="csv")
        # Here we will miss a lot of CSV files, because we only wrote parquet:
        assert len(vens.keys()) > len(fromparquetdisk2.keys())

        fromcsvdisk2 = VirtualEnsemble()
        fromcsvdisk2.from_disk("vens_dumped_csv", fmt="parquet")
        # But even if we only try to load parquet files, when CSV
        # files are found without corresponding parquet, the CSV file
        # will be read.
        assert set(vens.keys()) == set(fromcsvdisk2.keys())

    # Test manual intervention:
    fooframe = pd.DataFrame(data=np.random.randn(3, 3), columns=["FOO", "BAR", "COM"])
    fooframe.to_csv(os.path.join("vens_dumped", "share/results/tables/randomdata.csv"))
    manualens = VirtualEnsemble(fromdisk="vens_dumped")
    assert "share/results/tables/randomdata.csv" not in manualens.keys()

    # Now with correct column header,
    # but floating point data for realizations..
    fooframe = pd.DataFrame(data=np.random.randn(3, 3), columns=["REAL", "BAR", "COM"])
    fooframe.to_csv(os.path.join("vens_dumped", "share/results/tables/randomdata.csv"))
    manualens = VirtualEnsemble(fromdisk="vens_dumped")
    assert "share/results/tables/randomdata.csv" not in manualens.keys()

    # Now with correct column header, and with integer data for REAL..
    fooframe = pd.DataFrame(
        data=np.random.randint(low=0, high=100, size=(3, 3)),
        columns=["REAL", "BAR", "COM"],
    )
    fooframe.to_csv(os.path.join("vens_dumped", "share/results/tables/randomdata.csv"))
    manualens = VirtualEnsemble(fromdisk="vens_dumped")
    assert "share/results/tables/randomdata.csv" in manualens.keys()


def test_todisk_includefile(tmpdir):
    """Test that we can write VirtualEnsembles to the filesystem in a
    retrievable manner with discovered files included"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")
    reekensemble = ScratchEnsemble(
        "reektest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )

    tmpdir.chdir()

    reekensemble.load_smry(time_index="monthly", column_keys="*")
    reekensemble.load_smry(time_index="daily", column_keys="*")
    reekensemble.load_smry(time_index="yearly", column_keys="F*")
    reekensemble.load_smry(column_keys="FOPT")

    reekensemble.load_scalar("npv.txt")
    reekensemble.load_txt("outputs.txt")
    vens = reekensemble.to_virtual()

    vens.to_disk("vens_dumped_files", delete=True, includefiles=True, symlinks=True)
    for real in [0, 1, 2, 4, 4]:
        runpath = os.path.join(
            "vens_dumped_files", "__discoveredfiles", "realization-" + str(real)
        )
        assert os.path.exists(runpath)
        assert os.path.exists(
            os.path.join(runpath, "eclipse/model/2_R001_REEK-" + str(real) + ".UNSMRY")
        )


def test_get_smry_meta(tmpdir):
    """Test the conservation of smry meta-data in virtual ensembles"""

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    reekensemble = ScratchEnsemble(
        "reekmetatest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    # If no smry loaded before virtualization, nothing should be there:
    assert "__smry_metadata" not in reekensemble.to_virtual().keys()

    reekensemble.load_smry(time_index="yearly", column_keys=["F*"])
    origmeta = reekensemble.get_smry_meta()
    vens = reekensemble.to_virtual()
    assert "__smry_metadata" in vens.keys()
    meta = vens.get_df("__smry_metadata")
    # Internally it is stored as a DataFrame, we check that
    # since it is possible to get it using get_df(), and thereby
    # almost part of the API
    assert isinstance(meta, pd.DataFrame)

    # But rather users should use get_smry_meta() to obtain
    # stuff from the internal frame __smry_metadata:
    metadict = vens.get_smry_meta()
    assert isinstance(metadict, dict)
    assert len(metadict) + 2 == len(
        vens.get_smry(time_index="yearly", column_keys="*").columns
    )
    # (the vens only knows of F* columns)
    assert len(metadict) + 2 == len(
        vens.get_smry(time_index="yearly", column_keys="F*").columns
    )

    assert origmeta["FOPT"] == metadict["FOPT"]
    assert origmeta["FWPTH"] == metadict["FWPTH"]

    assert not vens.get_smry_meta([])
    assert vens.get_smry_meta(column_keys="FOPT")["FOPT"] == origmeta["FOPT"]

    assert not vens.get_smry_meta(column_keys="WOPT:NOTEXISTING")

    # Test that it is retrievable after dumping to disk:
    vens_disk_path = str(tmpdir.join("vens_dumped"))
    vens.to_disk(vens_disk_path)
    disk_vens = VirtualEnsemble(fromdisk=vens_disk_path)
    metadict = disk_vens.get_smry_meta()
    assert isinstance(metadict, dict)
    assert len(metadict) + 2 == len(
        vens.get_smry(time_index="yearly", column_keys="*").columns
    )
    # (the vens only knows of F* columns)
    assert len(metadict) + 2 == len(
        vens.get_smry(time_index="yearly", column_keys="F*").columns
    )

    assert origmeta["FOPT"] == metadict["FOPT"]
    assert origmeta["FWPTH"] == metadict["FWPTH"]


def test_get_smry_interpolation():
    """Test the summary resampling code for virtual ensembles"""

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    reekensemble = ScratchEnsemble(
        "reektest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    reekensemble.load_smry(time_index="yearly", column_keys=["F*"])
    reekensemble.load_scalar("npv.txt")
    vens_yearly = reekensemble.to_virtual()
    reekensemble.load_smry(time_index="monthly", column_keys=["F*"])
    # Create a vens that contains both monthly and yearly:
    vens_monthly = reekensemble.to_virtual()
    assert "npv.txt" in vens_monthly.keys()
    reekensemble.load_smry(time_index="daily", column_keys=["F*"])
    _ = reekensemble.to_virtual()  # monthly, yearly *and* daily

    # Resample yearly to monthly:
    monthly = vens_yearly.get_smry(column_keys="FOPT", time_index="monthly")
    assert "FOPT" in monthly.columns
    assert "REAL" in monthly.columns
    assert "DATE" in monthly.columns
    assert len(monthly["REAL"].unique()) == 5

    # 12 months pr. year, including final 1. jan, four years, 5 realizations:
    assert len(monthly) == (12 * 4 + 1) * 5

    for realidx in monthly["REAL"].unique():
        int_m = monthly.set_index("REAL").loc[realidx].set_index("DATE")
        true_m = (
            reekensemble.get_smry(column_keys="FOPT", time_index="monthly")
            .set_index("REAL")
            .loc[realidx]
            .set_index("DATE")
        )
        difference = int_m["FOPT"] - true_m["FOPT"]

        # The interpolation error should be zero at each 1st of January
        # but most likely nonzero elsewhere (at least for these realization)
        assert difference.loc["2001-01-01"] < 0.0001
        assert abs(difference.loc["2001-06-01"]) > 0
        assert difference.loc["2002-01-01"] < 0.0001
        assert abs(difference.loc["2002-06-01"]) > 0
        assert difference.loc["2003-01-01"] < 0.0001

    daily = vens_yearly.get_smry(column_keys=["FOPT", "FOPR"], time_index="daily")
    assert "FOPT" in daily.columns
    assert "REAL" in daily.columns
    assert "DATE" in daily.columns
    assert len(daily["REAL"].unique()) == 5
    assert len(daily) == (365 * 4 + 2) * 5  # 2003-01-01 and 2003-01-02 at end

    # Linear interpolation will give almost unique values everywhere:
    assert len(daily["FOPT"].unique()) > (365 * 4) * 5
    # While bfill for rates cannot be more unique than the yearly input
    assert len(daily["FOPR"].unique()) < 4 * 5  # Must be less than the numbers


def test_volumetric_rates():
    """Test the summary resampling code for virtual ensembles

    We only need to test the aggregation here.
    """

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    reekensemble = ScratchEnsemble(
        "reektest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    reekensemble.load_smry(time_index="yearly", column_keys=["F*"])
    reekensemble.load_scalar("npv.txt")
    vens = reekensemble.to_virtual()

    vol_rates = vens.get_volumetric_rates(column_keys="FOPT", time_index="yearly")
    assert isinstance(vol_rates, pd.DataFrame)
    assert "REAL" in vol_rates
    assert "DATE" in vol_rates
    assert "FOPR" in vol_rates
    assert len(vol_rates) == 25


def test_get_df_merge():
    """Testing merge support in get_df()"""

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    reekensemble = ScratchEnsemble(
        "reektest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    reekensemble.load_smry(time_index="yearly", column_keys=["F*"])
    reekensemble.load_scalar("npv.txt")
    reekensemble.load_csv("share/results/volumes/simulator_volume_fipnum.csv")
    outputs = reekensemble.load_txt("outputs.txt")
    vens = reekensemble.to_virtual()

    params = vens.get_df("parameters.txt")
    smrycount = len(vens.get_df("unsmry--yearly").columns)
    smryparams = vens.get_df("unsmry--yearly", merge="parameters")

    # The "minus 1" is due to the REAL column being present in both tables.
    assert len(smryparams.columns) == len(params.columns) + smrycount - 1

    paramsoutputs = vens.get_df("parameters", merge=["outputs"])
    assert len(paramsoutputs.columns) == len(params.columns) + len(outputs.columns) - 1

    assert (
        len(vens.get_df("unsmry--yearly", merge=["parameters", "outputs"]).columns)
        == smrycount + len(params.columns) + len(outputs.columns) - 2
    )

    assert (
        len(vens.get_df("parameters", merge="npv.txt").columns)
        == len(params.columns) + 1
    )
    # Symmetry:
    assert (
        len(vens.get_df("npv.txt", merge="parameters.txt").columns)
        == len(params.columns) + 1
    )

    # Merge with zone data, inject a mocked dataframe to the realization:
    vens.data["fipnum2zone"] = pd.DataFrame(
        columns=["FIPNUM", "ZONE"],
        data=[
            [1, "UpperReek"],
            [2, "MidReek"],
            [3, "LowerReek"],
            [4, "UpperReek"],
            [5, "MidReek"],
            [6, "LowerReek"],
        ],
    )
    volframe = vens.get_df("simulator_volume_fipnum", merge="fipnum2zone")
    assert "ZONE" in volframe
    assert "FIPNUM" in volframe
    assert "STOIIP_OIL" in volframe
    assert len(volframe["ZONE"].unique()) == 3
